import logging
import functools

from collections import OrderedDict

from qtpy.QtCore import (QObject, Slot, Signal, Property,
                         Q_ENUMS, QSize)
from qtpy.QtWidgets import (QWidget, QPlainTextEdit, QComboBox, QLabel,
                            QPushButton, QHBoxLayout, QVBoxLayout,
                            QStyleOption, QStyle)
from qtpy.QtGui import QPainter

logger = logging.getLogger(__name__)


def logger_destroyed(log, obj):
    """
    Callback invoked when the Widget is destroyed.
    This method is used to ensure that the log handlers are cleared.

    Parameters
    ----------
    log : Logger
        The logger object being used by the PyDMLogDisplay widget.
    obj : QWidget
        The widget which is being destroyed. All childs of this widget are
        already destroyed and must not be accessed.
    """
    if log:
        for handler in log.handlers:
            log.removeHandler(handler)


class GuiHandler(QObject, logging.Handler):
    """
    Handler for PyDM Applications

    A composite of a QObject and a logging handler. This can be added to a
    ``logging.Logger`` object just like any standard ``logging.Handler`` and
    will emit logging messages as Signals

    .. code:: python

        # Create a log and GuiHandler
        logger = logging.getLogger()
        ui_handler = GuiHandler(level=logging.INFO)
        # Attach our handler to the log
        logger.addHandler(ui_handler)
        # Publish log message via Signal
        ui_handler.message.connect(mySlot)

    Parameters
    ----------
    level: int
        Level of Handler

    parent: QObject, optional
    """
    message = Signal(str)

    def __init__(self, level=logging.NOTSET, parent=None):
        logging.Handler.__init__(self, level=level)
        QObject.__init__(self, parent=parent)

    def emit(self, record):
        """Emit formatted log messages when received but only if level is set."""
        # Avoid garbage to be presented when master log is running with DEBUG.
        if self.level == logging.NOTSET:
            return
        try:
            self.message.emit(self.format(record))
        except RuntimeError:
            logger.debug('Handler was destroyed at the C++ level.')


class LogLevels(object):
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50

    @staticmethod
    def as_dict():
        """
        Returns an ordered dict of LogLevels ordered by value.

        Returns
        -------
        OrderedDict
        """
        # First let's remove the internals
        entries = [(k, v) for k, v in LogLevels.__dict__.items() if
                   not k.startswith("__") and not callable(v) and not isinstance(v, staticmethod)]

        return OrderedDict(sorted(entries, key=lambda x: x[1], reverse=False))


class PyDMLogDisplay(QWidget, LogLevels):
    """
    Standard display for Log Output

    This widget handles instantating a ``GuiHandler`` and displaying log
    messages to a ``QPlainTextEdit``. The level of the log can be changed from
    inside the widget itself, allowing users to select from any of the
    ``.levels`` specified by the widget.

    Parameters
    ----------
    parent : QObject, optional

    logname : str
        Name of log to display in widget

    level : logging.Level
        Initial level of log display

    """
    Q_ENUMS(LogLevels)
    LogLevels = LogLevels
    terminator = '\n'
    default_format = '%(asctime)s %(message)s'
    default_level = logging.INFO

    def __init__(self, parent=None, logname=None, level=logging.NOTSET):
        QWidget.__init__(self, parent=parent)
        # Create Widgets
        self.label = QLabel('Minimum displayed log level: ', parent=self)
        self.combo = QComboBox(parent=self)
        self.text = QPlainTextEdit(parent=self)
        self.text.setReadOnly(True)
        self.clear_btn = QPushButton("Clear", parent=self)
        # Create layout
        layout = QVBoxLayout()
        level_control = QHBoxLayout()
        level_control.addWidget(self.label)
        level_control.addWidget(self.combo)
        layout.addLayout(level_control)
        layout.addWidget(self.text)
        layout.addWidget(self.clear_btn)
        self.setLayout(layout)
        # Allow QCombobox to control log level
        for log_level, value in LogLevels.as_dict().items():
            self.combo.addItem(log_level, value)
        self.combo.currentIndexChanged[str].connect(self.setLevel)
        # Allow QPushButton to clear log text
        self.clear_btn.clicked.connect(self.clear)
        # Create a handler with the default format
        self.handler = GuiHandler(level=level, parent=self)
        self.logFormat = self.default_format
        self.handler.message.connect(self.write)
        # Create logger. Either as a root or given logname
        self.log = None
        self.level = None
        self.logName = logname or ''
        self.logLevel = level
        self.destroyed.connect(functools.partial(logger_destroyed, self.log))

    def sizeHint(self):
        return QSize(400, 300)

    @Property(LogLevels)
    def logLevel(self):
        return self.level

    @logLevel.setter
    def logLevel(self, level):
        if level != self.level:
            self.level = level
            idx = self.combo.findData(level)
            self.combo.setCurrentIndex(idx)

    @Property(str)
    def logName(self):
        """Name of associated log"""
        return self.log.name

    @logName.setter
    def logName(self, name):
        # Disconnect prior log from handler
        if self.log:
            self.log.removeHandler(self.handler)
        # Reattach handler to new handler
        self.log = logging.getLogger(name)
        # Ensure that the log matches level of handler
        # only if the handler level is less than the log.
        if self.log.level < self.handler.level:
            self.log.setLevel(self.handler.level)
        # Attach preconfigured handler
        self.log.addHandler(self.handler)

    @Property(str)
    def logFormat(self):
        """Format for log messages"""
        return self.handler.formatter._fmt

    @logFormat.setter
    def logFormat(self, fmt):
        self.handler.setFormatter(logging.Formatter(fmt))

    @Slot(str)
    def write(self, message):
        """Write a message to the log display"""
        # We split the incoming message by new lines. In prior iterations of
        # this widget it was discovered that large blocks of text cause issues
        # at the Qt level.
        for msg in message.split(self.terminator):
            self.text.appendPlainText(msg)

    @Slot()
    def clear(self):
        """Clear the text area."""
        self.text.clear()

    @Slot(str)
    def setLevel(self, level):
        """Set the level of the contained logger"""
        # Get the level from the incoming string specification
        try:
            level = getattr(logging, level.upper())
        except AttributeError as exc:
            logger.exception("Invalid logging level specified %s",
                             level.upper())
        else:
            # Set the existing handler and logger to this level
            self.handler.setLevel(level)
            if self.log.level > self.handler.level or self.log.level == logging.NOTSET:
                self.log.setLevel(self.handler.level)

    def paintEvent(self, _):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        At PyDMDrawing this method handles the alarm painting with parameters
        from the stylesheet, configures the brush, pen and calls ```draw_item```
        so the specifics can be performed for each of the drawing classes.

        Parameters
        ----------
        event : QPaintEvent
        """
        painter = QPainter(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)



import os
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication
from ...widgets import PyDMSlider, PyDMTemplateRepeater

test_template_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "../test_data", "template.ui")
    
def test_template_file(qtbot):
    # Test that loading a template and setting data instantiates an instance
    # of the template.
    template_repeater = PyDMTemplateRepeater()
    qtbot.addWidget(template_repeater)
    template_repeater.templateFilename = test_template_path
    test_data = [{"devname": "test_device"}]
    template_repeater.data = test_data
    assert template_repeater.count() == len(test_data)
    slider = template_repeater.findChild(PyDMSlider, "bCtrlSlider")
    assert slider is not None
    assert slider.channel == "ca://{}:BCTRL".format(test_data[0]["devname"])


import logging

from qtpy.QtCore import (Qt, QSize, Property, Slot, Q_ENUMS, QMargins)
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import (QWidget, QButtonGroup, QGridLayout, QPushButton,
                            QRadioButton, QStyleOption, QStyle)

from .base import PyDMWritableWidget
from .. import data_plugins


class WidgetType(object):
    PushButton = 0
    RadioButton = 1


class_for_type = [QPushButton, QRadioButton]

logger = logging.getLogger(__name__)

class PyDMEnumButton(QWidget, PyDMWritableWidget, WidgetType):
    """
    A QWidget that renders buttons for every option of Enum Items.
    For now, two types of buttons can be rendered:
    - Push Button
    - Radio Button

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.

    Signals
    -------
    send_value_signal : int, float, str, bool or np.ndarray
        Emitted when the user changes the value.
    """
    Q_ENUMS(WidgetType)
    WidgetType = WidgetType

    def __init__(self, parent=None, init_channel=None):
        QWidget.__init__(self, parent)
        PyDMWritableWidget.__init__(self, init_channel=init_channel)
        self._invert_order = False
        self._use_custom_order = False
        self._custom_order = []
        self._has_enums = False
        self._checkable = True
        self.setLayout(QGridLayout(self))
        self._layout_spacing_horizontal = 6
        self._layout_spacing_vertical = 6
        self._layout_margins = QMargins(9, 9, 9, 9)
        self._btn_group = QButtonGroup()
        self._btn_group.setExclusive(True)
        self._btn_group.buttonClicked[int].connect(self.handle_button_clicked)
        self._widget_type = WidgetType.PushButton
        self._orientation = Qt.Vertical
        self._widgets = []
        self.rebuild_widgets()

    def minimumSizeHint(self):
        """
        This property holds the recommended minimum size for the widget.

        Returns
        -------
        QSize
        """
        # This is totally arbitrary, I just want *some* visible nonzero size
        return QSize(50, 100)

    @Property("QStringList")
    def items(self):
        """
        Items to be displayed in the button group.

        This property can be overridden by the items coming from the control system.
        Because C++ QStringList expects a list type, we need to make sure that None is never returned.

        Returns
        -------
        List[str]
        """
        return self.enum_strings or []

    @items.setter
    def items(self, value):
        self.enum_strings_changed(value)

    @Property(bool)
    def useCustomOrder(self):
        """
        Whether or not to use custom order for the button group.

        Returns
        -------
        bool
        """
        return self._use_custom_order

    @useCustomOrder.setter
    def useCustomOrder(self, value):
        if value != self._use_custom_order:
            self._use_custom_order = value
            self.rebuild_layout()

    @Property(bool)
    def invertOrder(self):
        """
        Whether or not to invert the order for the button group.

        Returns
        -------
        bool
        """
        return self._invert_order

    @invertOrder.setter
    def invertOrder(self, value):
        if value != self._invert_order:
            self._invert_order = value
            if self._has_enums:
                self.rebuild_layout()

    @Property("QStringList")
    def customOrder(self):
        """
        Index list in which items are to be displayed in the button group.

        Returns
        -------
        List[str]
        """
        return self._custom_order

    @customOrder.setter
    def customOrder(self, value):
        if value != self._custom_order:
            try:
                v = [int(v) for v in value]
            except ValueError:
                logger.error('customOrder values can only be integers.')
                return
            self._custom_order = value
            if self.useCustomOrder and self._has_enums:
                self.rebuild_layout()

    @Property(WidgetType)
    def widgetType(self):
        """
        The widget type to be used when composing the group.

        Returns
        -------
        WidgetType
        """
        return self._widget_type

    @widgetType.setter
    def widgetType(self, new_type):
        """
        The widget type to be used when composing the group.

        Parameters
        ----------
        new_type : WidgetType
        """
        if new_type != self._widget_type:
            self._widget_type = new_type
            self.rebuild_widgets()

    @Property(Qt.Orientation)
    def orientation(self):
        """
        Whether to lay out the bit indicators vertically or horizontally.

        Returns
        -------
        int
        """
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation):
        """
        Whether to lay out the bit indicators vertically or horizontally.

        Parameters
        -------
        new_orientation : Qt.Orientation, int
        """
        if new_orientation != self._orientation:
            self._orientation = new_orientation
            self.rebuild_layout()

    @Property(int)
    def marginTop(self):
        """
        The top margin of the QGridLayout of buttons.

        Returns
        -------
        int
        """
        return self._layout_margins.top()

    @marginTop.setter
    def marginTop(self, new_margin):
        """
        Set the top margin of the QGridLayout of buttons.

        Parameters
        -------
        int
        """
        new_margin = max(0, int(new_margin))
        self._layout_margins.setTop(new_margin)
        self.layout().setContentsMargins(self._layout_margins)

    @Property(int)
    def marginBottom(self):
        """
        The bottom margin of the QGridLayout of buttons.

        Returns
        -------
        int
        """
        return self._layout_margins.bottom()

    @marginBottom.setter
    def marginBottom(self, new_margin):
        """
        Set the bottom margin of the QGridLayout of buttons.

        Parameters
        -------
        int
        """
        new_margin = max(0, int(new_margin))
        self._layout_margins.setBottom(new_margin)
        self.layout().setContentsMargins(self._layout_margins)

    @Property(int)
    def marginLeft(self):
        """
        The left margin of the QGridLayout of buttons.

        Returns
        -------
        int
        """
        return self._layout_margins.left()

    @marginLeft.setter
    def marginLeft(self, new_margin):
        """
        Set the left margin of the QGridLayout of buttons.

        Parameters
        -------
        int
        """
        new_margin = max(0, int(new_margin))
        self._layout_margins.setLeft(new_margin)
        self.layout().setContentsMargins(self._layout_margins)

    @Property(int)
    def marginRight(self):
        """
        The right margin of the QGridLayout of buttons.

        Returns
        -------
        int
        """
        return self._layout_margins.right()

    @marginRight.setter
    def marginRight(self, new_margin):
        """
        Set the right margin of the QGridLayout of buttons.

        Parameters
        -------
        int
        """
        new_margin = max(0, int(new_margin))
        self._layout_margins.setRight(new_margin)
        self.layout().setContentsMargins(self._layout_margins)

    @Property(int)
    def horizontalSpacing(self):
        """
        The horizontal gap of the QGridLayout containing the QButtonGroup.

        Returns
        -------
        int
        """
        return self._layout_spacing_horizontal

    @horizontalSpacing.setter
    def horizontalSpacing(self, new_spacing):
        """
        Set the layout horizontal gap between buttons.

        Parameters
        -------
        new_spacing : int
        """
        new_spacing = max(0, int(new_spacing))
        if new_spacing != self._layout_spacing_horizontal:
            self._layout_spacing_horizontal = new_spacing
            self.layout().setHorizontalSpacing(new_spacing)

    @Property(int)
    def verticalSpacing(self):
        """
        The vertical gap of the QGridLayout containing the QButtonGroup.

        Returns
        -------
        int
        """
        return self._layout_spacing_vertical

    @verticalSpacing.setter
    def verticalSpacing(self, new_spacing):
        """
        Set the layout vertical gap between buttons.

        Parameters
        -------
        new_spacing : int
        """
        new_spacing = max(0, int(new_spacing))
        if new_spacing != self._layout_spacing_vertical:
            self._layout_spacing_vertical = new_spacing
            self.layout().setVerticalSpacing(new_spacing)

    @Property(bool)
    def checkable(self):
        """
        Whether or not the button should be checkable.

        Returns
        -------
        bool
        """
        return self._checkable

    @checkable.setter
    def checkable(self, value):
        if value != self._checkable:
            self._checkable = value
            for widget in self._widgets:
                widget.setCheckable(value)

    @Slot(int)
    def handle_button_clicked(self, id):
        """
        Handles the event of a button being clicked.

        Parameters
        ----------
        id : int
            The clicked button id.
        """
        self.send_value_signal.emit(id)

    def clear(self):
        """
        Remove all inner widgets from the layout
        """
        for col in range(0, self.layout().columnCount()):
            for row in range(0, self.layout().rowCount()):
                item = self.layout().itemAtPosition(row, col)
                if item is not None:
                    w = item.widget()
                    if w is not None:
                        w.hide()
                        self.layout().removeWidget(w)

    def rebuild_widgets(self):
        """
        Rebuild the list of widgets based on a new enum or generates a default
        list of fake strings so we can see something at Designer.
        """

        def generate_widgets(items):
            while len(self._widgets) != 0:
                w = self._widgets.pop(0)
                w.hide()
                self._btn_group.removeButton(w)
                w.deleteLater()

            for idx, entry in enumerate(items):
                w = class_for_type[self._widget_type](parent=self)
                w.setCheckable(self.checkable)
                w.setText(entry)
                w.setVisible(False)
                self._widgets.append(w)
                self._btn_group.addButton(w, idx)

        self.clear()
        if self._has_enums:
            generate_widgets(self.enum_strings)
        else:
            generate_widgets(["Item 1", "Item 2", "Item ..."])

        self.rebuild_layout()

    def rebuild_layout(self):
        """
        Method to reorganize the top-level widget and its contents
        according to the layout property values.
        """
        self.clear()

        if self.useCustomOrder:
            order = [int(v) for v in self.customOrder]
        else:
            order = list(range(len(self._widgets)))

        if self.invertOrder:
            order = order[::-1]

        for i, idx in enumerate(order):
            try:
                widget = self._widgets[idx]
                widget.setVisible(True)
            except IndexError:
                if self._has_enums:
                    logger.error(
                        'Invalid index for PyDMEnumButton %s. Index: %s, Range: 0 to %s',
                        self.objectName(), idx, len(self._widgets) - 1)
                continue
            if self.orientation == Qt.Vertical:
                self.layout().addWidget(widget, i, 0)
            elif self.orientation == Qt.Horizontal:
                self.layout().addWidget(widget, 0, i)

    def check_enable_state(self):
        """
        Checks whether or not the widget should be disable.
        This method also disables the widget and add a Tool Tip
        with the reason why it is disabled.

        """
        status = self._write_access and self._connected and self._has_enums
        tooltip = ""
        if not self._connected:
            tooltip += "Channel is disconnected."
        elif not self._write_access:
            if data_plugins.is_read_only():
                tooltip += "Running PyDM on Read-Only mode."
            else:
                tooltip += "Access denied by Channel Access Security."
        elif not self._has_enums:
            tooltip += "Enums not available."

        self.setToolTip(tooltip)
        self.setEnabled(status)

    def value_changed(self, new_val):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : int
            The new value from the channel.
        """
        if new_val is not None and new_val != self.value:
            super(PyDMEnumButton, self).value_changed(new_val)
            btn = self._btn_group.button(new_val)
            if btn:
                btn.setChecked(True)

    def enum_strings_changed(self, new_enum_strings):
        """
        Callback invoked when the Channel has new enum values.
        This callback also triggers a value_changed call so the
        new enum values to be broadcasted.

        Parameters
        ----------
        new_enum_strings : tuple
            The new list of values
        """
        if new_enum_strings is not None \
                and new_enum_strings != self.enum_strings:
            super(PyDMEnumButton, self).enum_strings_changed(new_enum_strings)
            self._has_enums = True
            self.check_enable_state()
            self.rebuild_widgets()

    def paintEvent(self, _):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        At PyDMDrawing this method handles the alarm painting with parameters
        from the stylesheet, configures the brush, pen and calls ```draw_item```
        so the specifics can be performed for each of the drawing classes.

        Parameters
        ----------
        event : QPaintEvent
        """
        painter = QPainter(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)



import warnings
warnings.warn("The display_module was renamed to display."
              "Please modify your code to reflect this change.",
              DeprecationWarning, stacklevel=2)

__all__ = ['Display', 'ScreenTarget',
           'load_file', 'load_py_file', 'load_ui_file']

from pydm.display import (Display, ScreenTarget,
                          load_file, load_py_file, load_ui_file)



import os
import platform
import weakref
import logging
import functools
import json
import numpy as np
from qtpy.QtWidgets import (QApplication, QMenu, QGraphicsOpacityEffect,
                            QToolTip, QWidget)
from qtpy.QtGui import QCursor, QIcon
from qtpy.QtCore import Qt, QEvent, Signal, Slot, Property
from .channel import PyDMChannel
from .. import data_plugins, tools, config
from ..utilities import is_qt_designer, remove_protocol
from ..display import Display
from .rules import RulesDispatcher

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

logger = logging.getLogger(__name__)

try:
    str_types = (str, unicode)
except NameError:
    str_types = (str,)


def get_icon_file(name):
    """
    Returns the full path to the icon represented by name.

    Parameters
    ----------
    name : str
        The filename to load the file path.

    Returns
    -------
    str
    """
    base_path = os.path.dirname(os.path.realpath(__file__))
    icon_path = os.path.join(base_path, "icons", "terminator.png")
    return icon_path


def is_channel_valid(channel):
    """
    Verify if a channel string is valid.
    For now a valid channel is a channel in which:
    - It is not None
    - It is not an empty string

    Parameters
    ----------
    channel : str
        The channel value

    Returns
    -------
    valid : bool
        Returns True if the channel is valid, False otherwise.
    """
    return channel is not None and channel != ""


def only_if_channel_set(fcn):
    """Decorator to avoid executing a method if a channel is not valid or configured."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        if is_channel_valid(self._channel):
            return fcn(self, *args, **kwargs)
        else:
            return

    return wrapper


def widget_destroyed(channels, widget):
    """
    Callback invoked when the Widget is destroyed.
    This method is used to ensure that the channels are disconnected.

    Parameters
    ----------
    channels : list
        A list of PyDMChannel objects that this widget uses.
    widget : weakref
        Weakref to the widget.
    """
    chs = channels()
    if chs:
        for ch in chs:
            if ch:
                ch.disconnect(destroying=True)

    RulesDispatcher().unregister(widget)


def refresh_style(widget):
    """
    Method that traverse the widget tree starting at `widget` and refresh the
    style for this widget and its childs.

    Parameters
    ----------
    widget : QWidget
    """
    widgets = [widget]

    try:
        widgets.extend(widget.findChildren(QWidget))
    except:
        # If we fail it means that widget is probably destroyed
        return
    for child_widget in widgets:
        try:
            child_widget.style().unpolish(child_widget)
            child_widget.style().polish(child_widget)
            child_widget.update()
        except Exception as ex:
            # Widget was probably destroyed
            logger.debug('Error while refreshing stylesheet. %s ', ex)


class PyDMPrimitiveWidget(object):
    """
    Primitive class that determines that a given widget is a PyDMWidget.
    All Widget classes from PyDMWidget will be True for
    isinstance(obj, PyDMPrimitiveWidget)
    """
    DEFAULT_RULE_PROPERTY = "Visible"
    RULE_PROPERTIES = {
        'Enable': ['setEnabled', bool],
        'Visible': ['setVisible', bool],
        'Opacity': ['set_opacity', float]
    }

    designer_icon = QIcon()

    def __init__(self, **kwargs):
        self.app = QApplication.instance()
        self._rules = None
        self._opacity = 1.0
        if not is_qt_designer():
            # We should  install the Event Filter only if we are running
            # and not at the Designer
            self.installEventFilter(self)

    def eventFilter(self, obj, event):
        """
        EventFilter to redirect "middle click" to :meth:`.show_address_tooltip`
        """
        # Override the eventFilter to capture all middle mouse button events,
        # and show a tooltip if needed.
        if event.type() == QEvent.MouseButtonPress:
            if event.button() == Qt.MiddleButton:
                self.show_address_tooltip(event)
                return True
        return False

    def show_address_tooltip(self, event):
        """
        Show the PyDMTooltip and copy address to clipboard

        This is intended to replicate the behavior of the "middle click" from
        EDM. If the QWidget does not have a valid PyDMChannel nothing will be
        displayed
        """
        channels_method = getattr(self, 'channels', None)
        if channels_method is None:
            return
        channels = channels_method()
        if not channels:
            logger.debug('Widget has no channels to display tooltip')
            return

        addrs = []
        no_proto_addrs = []
        for ch in channels:
            addr = ch.address
            if not addr:
                continue
            addrs.append(addr)
            no_proto_addrs.append(remove_protocol(addr))

        tooltip = os.linesep.join(addrs)
        clipboard_text = " ".join(no_proto_addrs)
        QToolTip.showText(event.globalPos(), tooltip)
        # If the address has a protocol, strip it out before putting it on the
        # clipboard.

        clipboard = QApplication.clipboard()

        mode = clipboard.Clipboard
        if platform.system() == 'Linux':
            # Mode Selection is only valid for X11.
            mode = clipboard.Selection

        clipboard.setText(clipboard_text, mode=mode)
        event = QEvent(QEvent.Clipboard)
        self.app.sendEvent(clipboard, event)

    def opacity(self):
        """
        Float value between 0 and 1 representing the opacity of the widget
        where 0 means transparent.

        Returns
        -------
        opacity : float
        """
        return self._opacity

    def set_opacity(self, val):
        """
        Float value between 0 and 1 representing the opacity of the widget
        where 0 means transparent.

        Parameters
        ----------
        val : float
            The new value for the opacity
        """
        op = QGraphicsOpacityEffect(self)
        if val > 1:
            val = 1
        elif val < 0:
            val = 0
        self._opacity = val
        op.setOpacity(val)  # 0 to 1 will cause the fade effect to kick in
        self.setGraphicsEffect(op)
        self.setAutoFillBackground(True)

    @Slot(dict)
    def rule_evaluated(self, payload):
        """
        Callback called when a rule has a new value for a property.

        Parameters
        ----------
        payload : dict
            Dictionary containing the rule name, the property to be set and the
            new value.

        Returns
        -------
        None
        """
        name = payload.get('name', '')
        prop = payload.get('property', '')
        value = payload.get('value', None)

        if prop not in self.RULE_PROPERTIES:
            logger.error('Error at Rule: %s. %s is not part of this widget properties.',
                         name, prop)
            return

        method_name, data_type = self.RULE_PROPERTIES[prop]
        try:
            if data_type == bool and isinstance(value, str_types):
                # We do this as we already import json and for Python:
                # bool("False") -> True
                val = json.loads(value.lower())
            else:
                val = data_type(value)

            method = getattr(self, method_name)
            if callable(method):
                method(val)
            else:
                setattr(self, method_name, val)

        except:
            logger.error('Error at Rule: %s. Could not execute method %s with '
                         'value %s and type as %s.',
                         name, method_name, value, data_type.__name__)

    @Property(str, designable=False)
    def rules(self):
        """
        JSON-formatted list of dictionaries, with rules for the widget.

        Returns
        -------
        str
        """
        return self._rules

    @rules.setter
    def rules(self, new_rules):
        """
        JSON-formatted list of dictionaries, with rules for the widget.

        Parameters
        ----------
        new_rules : str

        Returns
        -------
        None
        """
        if new_rules != self._rules:
            self._rules = new_rules
            RulesDispatcher().unregister(self)
            try:
                rules_list = json.loads(self._rules)
                if rules_list:
                    RulesDispatcher().register(self, rules_list)
            except JSONDecodeError as ex:
                logger.exception('Invalid format for Rules')

    def find_parent_display(self):
        widget = self.parent()
        while widget is not None:
            if isinstance(widget, Display):
                return widget
            widget = widget.parent()
        return None


class TextFormatter(object):

    default_precision_from_pv = True

    def __init__(self):
        self._show_units = False
        self.format_string = "{}"
        self._precision_from_pv = None
        self._user_prec = 0
        self._prec = 0
        self._unit = ""

    def update_format_string(self):
        """
        Reconstruct the format string to be used when representing the
        output value.

        Returns
        -------
        format_string : str
            The format string to be used including or not the precision
            and unit
        """
        self.format_string = "{}"
        if isinstance(self.value, (int, float)):
            self.format_string = "{:." + str(self.precision) + "f}"
        if self._show_units and self._unit != "":
            self.format_string += " {}".format(self._unit)
        return self.format_string

    def precision_changed(self, new_precision):
        """
        Callback invoked when the Channel has new precision value.
        This callback also triggers an update_format_string call so the
        new precision value is considered.

        Parameters
        ----------
        new_precison : int or float
            The new precision value
        """
        if self.precisionFromPV and new_precision != self._prec:
            self._prec = new_precision
            if self.value is not None:
                self.value_changed(self.value)

    @Slot(int)
    @Slot(float)
    def precisionChanged(self, new_prec):
        """
        PyQT Slot for changes on the precision of the Channel
        This slot sends the new precision value to the
        ```precision_changed``` callback.

        Parameters
        ----------
        new_prec : int or float
        """
        self.precision_changed(new_prec)

    @Property(int)
    def precision(self):
        """
        The precision to be used when formatting the output of the PV

        Returns
        -------
        prec : int
            The current precision value
        """
        if self.precisionFromPV:
            return self._prec
        return self._user_prec

    @precision.setter
    def precision(self, new_prec):
        """
        The precision to be used when formatting the output of the PV.
        This has no effect when ```precisionFromPV``` is True.

        Parameters
        ----------
        new_prec : int
            The new precision value to use
        """
        # Only allow one to change the property if not getting the precision
        # from the PV
        if self._precision_from_pv is not None and self._precision_from_pv:
            return
        if new_prec and self._prec != int(new_prec) and new_prec >= 0:
            self._user_prec = int(new_prec)
            if not is_qt_designer() or config.DESIGNER_ONLINE:
                self.value_changed(self.value)

    @Slot(str)
    def unitChanged(self, new_unit):
        """
        PyQT Slot for changes on the unit of the Channel
        This slot sends the new unit string to the
        ```unit_changed``` callback.

        Parameters
        ----------
        new_unit : str
        """
        self.unit_changed(new_unit)

    def unit_changed(self, new_unit):
        """
        Callback invoked when the Channel has new unit value.
        This callback also triggers an update_format_string call so the
        new unit value is considered if ```showUnits``` is set.

        Parameters
        ----------
        new_unit : str
            The new unit
        """
        if self._unit != new_unit:
            self._unit = new_unit
            if self.value is not None:
                self.value_changed(self.value)


    @Property(bool)
    def showUnits(self):
        """
        A choice whether or not to show the units given by the channel

        If set to True, the units given in the channel will be displayed
        with the value. If using an EPICS channel, this will automatically
        be linked to the EGU field of the PV.

        Returns
        -------
        show_units : bool
            True means that the unit will be appended to the output value
            format string
        """
        return self._show_units

    @showUnits.setter
    def showUnits(self, show_units):
        """
        A choice whether or not to show the units given by the channel

        If set to True, the units given in the channel will be displayed
        with the value. If using an EPICS channel, this will automatically
        be linked to the EGU field of the PV.

        Paramters
        ---------
        show_units : bool
            True means that the unit will be appended to the output value
            format string
        """
        if self._show_units != show_units:
            self._show_units = show_units
            self.update_format_string()

    @Property(bool)
    def precisionFromPV(self):
        """
        A choice whether or not to use the precision given by channel.

        If set to False, the value received will be displayed as is, with
        no modification to the number of displayed significant figures.
        However, if set to True, and the channel specifies a display
        precision, a float or integer channel value will be set to display
        the correct precision. When using an EPICS Channel, the precision
        value corresponds to the PV's PREC field.

        It is also important to note, that if the value of the channel
        is a String, the choice of True or False will have no affect on
        the display.

        Returns
        -------
        precison_from_pv : bool
            True means that the widget will use the precision information
            from the Channel if available.
        """
        return (self._precision_from_pv
                if self._precision_from_pv is not None
                else self.default_precision_from_pv)

    @precisionFromPV.setter
    def precisionFromPV(self, value):
        """
        A choice whether or not to use the precision given by channel.

        If set to False, the value received will be displayed as is, with
        no modification to the number of displayed significant figures.
        However, if set to True, and the channel specifies a display
        precision, a float or integer channel value will be set to
        display the correct precision. When using an EPICS Channel, the
        precision value corresponds to the PV's PREC field.

        It is also important to note, that if the value of the channel is
        a String, the choice of True or False will have no affect on the
        display.

        Parameters
        ----------
        value : bool
            True means that the widget will use the precision information
            from the PV if available.
        """
        if self._precision_from_pv is None or self._precision_from_pv != bool(value):
            self._precision_from_pv = value
            self.update_format_string()

    def value_changed(self, new_val):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : str, int, float, bool or np.ndarray
            The new value from the channel. The type depends on the channel.
        """
        super(TextFormatter, self).value_changed(new_val)
        self.update_format_string()


class PyDMWidget(PyDMPrimitiveWidget):
    """
    PyDM base class for Read-Only widgets.
    This class implements all the functions of connection, alarm
    handling and more.

    Parameters
    ----------
    init_channel : str, optional
        The channel to be used by the widget.

    """
    # Alarm types
    ALARM_NONE = 0
    ALARM_MINOR = 1
    ALARM_MAJOR = 2
    ALARM_INVALID = 3
    ALARM_DISCONNECTED = 4

    def __init__(self, init_channel=None):
        super(PyDMWidget, self).__init__()

        if not all([prop in PyDMPrimitiveWidget.RULE_PROPERTIES for prop in
                    ['Position - X', 'Position - Y']]):
            PyDMWidget.RULE_PROPERTIES = PyDMPrimitiveWidget.RULE_PROPERTIES.copy()
            PyDMWidget.RULE_PROPERTIES.update(
                {'Position - X': ['setX', int],
                 'Position - Y': ['setY', int]})

        self._connected = True
        self._channel = None
        self._channels = list()
        self._show_units = False
        self._alarm_sensitive_content = False
        self._alarm_sensitive_border = True
        self._alarm_state = self.ALARM_NONE
        self._tooltip = None

        self._upper_ctrl_limit = None
        self._lower_ctrl_limit = None

        self.enum_strings = None

        self.value = None
        self.channeltype = None
        self.subtype = None

        # If this label is inside a PyDMApplication (not Designer) start it in
        # the disconnected state.
        self.setContextMenuPolicy(Qt.DefaultContextMenu)
        self.contextMenuEvent = self.open_context_menu
        self.channel = init_channel
        if not is_qt_designer():
            self._connected = False
            self.alarmSeverityChanged(self.ALARM_DISCONNECTED)
            self.check_enable_state()

        self.destroyed.connect(
            functools.partial(widget_destroyed, self.channels, weakref.ref(self))
        )

    def widget_ctx_menu(self):
        """
        Fetch the Widget specific context menu which will be populated with additional tools by `assemble_tools_menu`.

        Returns
        -------
        QMenu or None
            If the return of this method is None a new QMenu will be created by `assemble_tools_menu`.
        """
        return None

    def generate_context_menu(self):
        """
        Generates the custom context menu, and populates it with any external
        tools that have been loaded.  PyDMWidget subclasses should override
        this method (after calling superclass implementation) to add the menu.

        Returns
        -------
        QMenu
        """
        menu = self.widget_ctx_menu()
        if menu is None:
            menu = QMenu(parent=self)

        kwargs = {'channels': self.channels_for_tools(), 'sender': self}
        tools.assemble_tools_menu(menu, widget_only=True, widget=self, **kwargs)
        return menu

    def open_context_menu(self, ev):
        """
        Handler for when the Default Context Menu is requested.

        Parameters
        ----------
        ev : QEvent
        """
        menu = self.generate_context_menu()
        action = menu.exec_(self.mapToGlobal(ev.pos()))
        del menu

    def init_for_designer(self):
        """
        Method called after the constructor to tweak configurations for
        when using the widget with the Qt Designer
        """
        self._connected = True

    def connection_changed(self, connected):
        """
        Callback invoked when the connection state of the Channel is changed.
        This callback acts on the connection state to enable/disable the widget
        and also trigger the change on alarm severity to ALARM_DISCONNECTED.

        Parameters
        ----------
        connected : int
            When this value is 0 the channel is disconnected, 1 otherwise.
        """
        self._connected = connected
        self.check_enable_state()
        if not connected:
            self.alarmSeverityChanged(self.ALARM_DISCONNECTED)
        else:
            self.alarmSeverityChanged(self.ALARM_NONE)

    def value_changed(self, new_val):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : str, int, float, bool or np.ndarray
            The new value from the channel. The type depends on the channel.
        """
        self.value = new_val
        self.channeltype = type(self.value)
        if self.channeltype == np.ndarray:
            self.subtype = self.value.dtype.type
        else:
            try:
                if self.channeltype == unicode:
                    # For Python 2.7, set the the channel type to str instead of unicode
                    self.channeltype = str
            except NameError:
                pass

    @Property(int, designable=False)
    def alarmSeverity(self):
        return self._alarm_state

    @alarmSeverity.setter
    def alarmSeverity(self, new_severity):
        if self._alarm_state != new_severity:
            self._alarm_state = new_severity

    def alarm_severity_changed(self, new_alarm_severity):
        """
        Callback invoked when the Channel alarm severity is changed.
        This callback is not processed if the widget has no channel
        associated with it.
        This callback handles the composition of the stylesheet to be
        applied and the call
        to update to redraw the widget with the needed changes for the
        new state.

        Parameters
        ----------
        new_alarm_severity : int
            The new severity where 0 = NO_ALARM, 1 = MINOR, 2 = MAJOR
            and 3 = INVALID
        """
        # 0 = NO_ALARM, 1 = MINOR, 2 = MAJOR, 3 = INVALID
        if new_alarm_severity == self._alarm_state:
            return
        if not self._channel:
            self._alarm_state = PyDMWidget.ALARM_NONE
        else:
            self._alarm_state = new_alarm_severity
        refresh_style(self)

    def enum_strings_changed(self, new_enum_strings):
        """
        Callback invoked when the Channel has new enum values.
        This callback also triggers a value_changed call so the
        new enum values to be broadcasted

        Parameters
        ----------
        new_enum_strings : tuple
            The new list of values
        """
        if new_enum_strings != self.enum_strings:
            self.enum_strings = new_enum_strings
            self.value_changed(self.value)

    def get_address(self):
        if not len(self._channels):
            logger.warning("Object %r has no PyDM Channels", self)
            return
        return self.channels()[0].address

    def ctrl_limit_changed(self, which, new_limit):
        """
        Callback invoked when the Channel receives new control limit
        values.

        Parameters
        ----------
        which : str
            Which control limit was changed. "UPPER" or "LOWER"
        new_limit : float
            New value for the control limit
        """
        if which == "UPPER":
            self._upper_ctrl_limit = new_limit
        else:
            self._lower_ctrl_limit = new_limit

    @Slot(bool)
    def connectionStateChanged(self, connected):
        """
        PyQT Slot for changes on the Connection State of the Channel
        This slot sends the connection state to the ```connection_changed```
        callback.

        Parameters
        ----------
        connected : bool
        """
        # false = disconnected, true = connected
        self.connection_changed(connected)

    @Slot(int)
    @Slot(float)
    @Slot(str)
    @Slot(bool)
    @Slot(np.ndarray)
    def channelValueChanged(self, new_val):
        """
        PyQT Slot for changes on the Value of the Channel
        This slot sends the value to the ```value_changed``` callback.

        Parameters
        ----------
        new_val : int, float, str, bool or np.ndarray
        """
        self.value_changed(new_val)

    @Slot(int)
    def alarmSeverityChanged(self, new_alarm_severity):
        """
        PyQT Slot for changes on the Alarm Severity of the Channel
        This slot sends the severity value to the
        ```alarm_severity_changed``` callback.

        Parameters
        ----------
        new_alarm_severity : int
        """
        self.alarm_severity_changed(new_alarm_severity)

    @Slot(tuple)
    def enumStringsChanged(self, new_enum_strings):
        """
        PyQT Slot for changes on the string values of the Channel
        This slot sends the new strings to the
        ```enum_strings_changed``` callback.

        Parameters
        ----------
        new_enum_strings : tuple
        """
        self.enum_strings_changed(new_enum_strings)

    @Slot(float)
    def upperCtrlLimitChanged(self, new_limit):
        """
        PyQT Slot for changes on the upper control limit value of the Channel
        This slot sends the new limit value to the
        ```ctrl_limit_changed``` callback.

        Parameters
        ----------
        new_limit : float
        """
        self.ctrl_limit_changed("UPPER", new_limit)

    @Slot(float)
    def lowerCtrlLimitChanged(self, new_limit):
        """
        PyQT Slot for changes on the lower control limit value of the Channel
        This slot sends the new limit value to the
        ```ctrl_limit_changed``` callback.

        Parameters
        ----------
        new_limit : float
        """
        self.ctrl_limit_changed("LOWER", new_limit)

    @Slot()
    def force_redraw(self):
        """
        PyQT Slot to force a redraw on the widget.

        """
        self.update()

    def setX(self, new_x):
        """
        Set the X position of the Widget on the screen.

        Parameters
        ----------
        new_x : int
            The new X position

        Returns
        -------
        None
        """
        point = self.pos()
        point.setX(new_x)
        self.move(point)

    def setY(self, new_y):
        """
        Set the Y position of the Widget on the screen.

        Parameters
        ----------
        new_y : int
            The new Y position

        Returns
        -------
        None
        """
        point = self.pos()
        point.setY(new_y)
        self.move(point)

    @Property(bool)
    def alarmSensitiveContent(self):
        """
        Whether or not the content color changes when alarm severity
        changes.

        Returns
        -------
        bool
            True means that the content color will be changed in case of
            alarm severity changes.
        """
        return self._alarm_sensitive_content

    @alarmSensitiveContent.setter
    def alarmSensitiveContent(self, checked):
        """
        Whether or not the content color changes when alarm severity
        changes.

        Parameters
        ----------
        checked : bool
            True means that the content color will be changed in case of
            alarm severity changes.
        """
        self._alarm_sensitive_content = checked
        self.alarm_severity_changed(self._alarm_state)

    @Property(bool)
    def alarmSensitiveBorder(self):
        """
        Whether or not the border color changes when alarm severity changes.

        Returns
        -------
        bool
            True means that the border color will be changed in case of
            alarm severity changes.
        """
        return self._alarm_sensitive_border

    @alarmSensitiveBorder.setter
    def alarmSensitiveBorder(self, checked):
        """
        Whether or not the border color changes when alarm severity
        changes.

        Parameters
        ----------
        checked : bool
            True means that the border color will be changed in case of
            alarm severity changes.
        """
        self._alarm_sensitive_border = checked
        self.alarm_severity_changed(self._alarm_state)

    @Property(str)
    def channel(self):
        """
        The channel address in use for this widget.

        Returns
        -------
        channel : str
            Channel address
        """
        if self._channel:
            return str(self._channel)
        return None

    @channel.setter
    def channel(self, value):
        """
        The channel address to use for this widget.

        Parameters
        ----------
        value : str
            Channel address
        """
        if self._channel != value:
            # Remove old connections
            for channel in [c for c in self._channels if
                            c.address == self._channel]:
                channel.disconnect()
                self._channels.remove(channel)
            # Load new channel
            self._channel = str(value)
            if not self._channel:
                logger.debug('Channel was set to an empty string.')
                return
            channel = PyDMChannel(address=self._channel,
                                  connection_slot=self.connectionStateChanged,
                                  value_slot=self.channelValueChanged,
                                  severity_slot=self.alarmSeverityChanged,
                                  enum_strings_slot=self.enumStringsChanged,
                                  unit_slot=None,
                                  prec_slot=None,
                                  upper_ctrl_limit_slot=self.upperCtrlLimitChanged,
                                  lower_ctrl_limit_slot=self.lowerCtrlLimitChanged,
                                  value_signal=None,
                                  write_access_slot=None)
            # Load writeable channels if our widget requires them. These should
            # not exist on the base PyDMWidget but prevents us from duplicating
            # the method below to only make two more connections
            if hasattr(self, 'writeAccessChanged'):
                channel.write_access_slot = self.writeAccessChanged
            if hasattr(self, 'send_value_signal'):
                channel.value_signal = self.send_value_signal
            # Do the same thing for classes that use the TextFormatter mixin.
            if hasattr(self, 'unitChanged'):
                channel.unit_slot = self.unitChanged
            if hasattr(self, 'precisionChanged'):
                channel.prec_slot = self.precisionChanged
            # Connect write channels if we have them
            channel.connect()
            self._channels.append(channel)

    def restore_original_tooltip(self):
        if self._tooltip is None:
            self._tooltip = self.toolTip()
        return self._tooltip

    @only_if_channel_set
    def check_enable_state(self):
        """
        Checks whether or not the widget should be disable.
        This method also disables the widget and add a Tool Tip
        with the reason why it is disabled.
        """
        status = self._connected
        tooltip = self.restore_original_tooltip()
        if not status:
            if tooltip != '':
                tooltip += '\n'
            tooltip += "PV is disconnected."
            tooltip += '\n'
            tooltip += self.get_address()

        self.setToolTip(tooltip)
        self.setEnabled(status)

    def get_ctrl_limits(self):
        """
        Returns a tuple with the control limits for the channel

        Returns
        -------
        (lower, upper) : tuple
            Lower and Upper control limits
        """
        return self._lower_ctrl_limit, self._upper_ctrl_limit

    def channels(self):
        """
        Returns the channels being used for this Widget.

        Returns
        -------
        channels : list
            List of PyDMChannel objects
        """
        if len(self._channels):
            return self._channels
        else:
            return None

    def channels_for_tools(self):
        """
        Returns a list of channels useful for external tools.
        The default implementation here is just to return
        self.channels(), but some widgets will want to re-implement
        this, especially if they have multiple channels, but only
        one real 'signal' channel.

        Returns
        -------
        list
        """
        return self.channels()


class PyDMWritableWidget(PyDMWidget):
    """
    PyDM base class for Writable widgets.
    This class implements the send_value_signal and also the event filter for write access changes on PVs.

    Parameters
    ----------
    init_channel : str, optional
        The channel to be used by the widget.

    Signals
    -------
    send_value_signal : int, float, str, bool or np.ndarray
        Emitted when the user changes the value
    """

    __Signals__ = ("send_value_signal([int], [float], [str], [bool], [np.ndarray])")

    # Emitted when the user changes the value.
    send_value_signal = Signal([int], [float], [str], [bool], [np.ndarray])

    def __init__(self, init_channel=None):
        self._write_access = False
        super(PyDMWritableWidget, self).__init__(init_channel=init_channel)

    def init_for_designer(self):
        """
        Method called after the constructor to tweak configurations for
        when using the widget with the Qt Designer
        """
        super(PyDMWritableWidget, self).init_for_designer()
        self._write_access = True

    def eventFilter(self, obj, event):
        """
        Filters events on this object.

        Params
        ------
        object : QObject
            The object that is being handled.
        event : QEvent
            The event that is happening.

        Returns
        -------
        bool
            True to stop the event from being handled further; otherwise
            return false.
        """
        channel = getattr(self, 'channel', None)
        if is_channel_valid(channel):
            status = self._write_access and self._connected

            if event.type() == QEvent.Leave:
                QApplication.restoreOverrideCursor()

            if event.type() == QEvent.Enter and not status:
                QApplication.setOverrideCursor(QCursor(Qt.ForbiddenCursor))

        return PyDMWidget.eventFilter(self, obj, event)

    def write_access_changed(self, new_write_access):
        """
        Callback invoked when the Channel has new write access value.
        This callback calls check_enable_state so it can act on the widget
        enabling or disabling it accordingly

        Parameters
        ----------
        new_write_access : bool
            True if write operations to the channel are allowed.
        """
        self._write_access = new_write_access
        self.check_enable_state()

    @Slot(bool)
    def writeAccessChanged(self, write_access):
        """
        PyQT Slot for changes on the write access value of the Channel
        This slot sends the write access value to the ```write_access_changed``` callback.

        Parameters
        ----------
        write_access : bool
        """
        self.write_access_changed(write_access)

    @only_if_channel_set
    def check_enable_state(self):
        """
        Checks whether or not the widget should be disable.
        This method also disables the widget and add a Tool Tip
        with the reason why it is disabled.
        """
        status = self._write_access and self._connected
        tooltip = self.restore_original_tooltip()
        if not self._connected:
            if tooltip != '':
                tooltip += '\n'
            tooltip += "PV is disconnected."
            tooltip += '\n'
            tooltip += self.get_address()
        elif not self._write_access:
            if tooltip != '':
                tooltip += '\n'
            if data_plugins.is_read_only():
                tooltip += "Running PyDM on Read-Only mode."
            else:
                tooltip += "Access denied by Channel Access Security."
        self.setToolTip(tooltip)
        self.setEnabled(status)



import collections
import functools
import logging
import math
import threading
import warnings

import numpy as np

try:
    from urllib import parse  # Python 3
except ImportError:
    import urlparse as parse

from qtpy.QtCore import Slot, QThread, Signal, Qt
from qtpy.QtWidgets import QApplication

import pydm
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection

logger = logging.getLogger(__name__)


def epics_string(value, string_encoding="utf-8"):
    # Stop at the first zero
    # Assume the ndarray is one-dimensional
    value = value.tobytes()
    try:
        value = value[:value.index(0)]
    except IndexError:
        pass
    return value.decode(string_encoding, "replace")  # <-- ignore decoding errors, just in case


class CalcThread(QThread):
    eval_env = {'math': math,
                'np': np,
                'numpy': np,
                'epics_string': epics_string}

    eval_env.update({k: v for k, v in math.__dict__.items() if k[0] != '_'})
    new_data_signal = Signal(dict)
    RESERVED_FIELD = ["update", "expr", "name"]

    def __init__(self, config, *args, **kwargs):
        QThread.__init__(self, *args, **kwargs)
        self.app = QApplication.instance()
        self.app.aboutToQuit.connect(self.requestInterruption)

        self.config = config
        self.listen_for_update = None

        self._calculate = threading.Event()
        self._names = []
        self._channels = []
        self._value = None
        self._values = collections.defaultdict(lambda: None)
        self._connections = collections.defaultdict(lambda: False)
        self._expression = self.config.get('expr', '')[0]

        channels = {}
        for key, channel in self.config.items():
            if key not in CalcThread.RESERVED_FIELD:
                channels[key] = channel[0]

        update = self.config.get('update', None)

        if update is not None:
            self.listen_for_update = update[0].split(',')
            self.listen_for_update = list(map(str.strip, self.listen_for_update))

        for name, channel in channels.items():
            conn_cb = functools.partial(self.callback_conn, name)
            value_cb = functools.partial(self.callback_value, name)
            c = pydm.PyDMChannel(channel, connection_slot=conn_cb,
                                 value_slot=value_cb)
            self._channels.append(c)
            self._names.append(name)

    @property
    def connected(self):
        return all(self._connections.values())

    def _connect(self):
        for ch in self._channels:
            ch.connect()

    def _disconnect(self):
        for ch in self._channels:
            ch.disconnect()

    def _send_update(self, conn, value):
        self.new_data_signal.emit({"connection": conn,
                                   "value": value})

    def run(self):
        self._connect()

        while True:
            self._calculate.wait()
            self._calculate.clear()
            if self.isInterruptionRequested():
                break
            self.calculate_expression()
        self._disconnect()

    def callback_value(self, name, value):
        """
        Callback executed when a channel receives a new value.

        Parameters
        ----------
        name : str
            The channel variable name.
        value : any
            The new value for this channel.

        Returns
        -------
        None
        """
        self._values[name] = value
        if not self.connected:
            logger.debug(
                "Calculation '%s': Not all channels are connected, skipping execution.",
                self.objectName())
            return

        if self.listen_for_update is None or name in self.listen_for_update:
            self._calculate.set()

    def callback_conn(self, name, value):
        """
        Callback executed when a channel connection status is changed.

        Parameters
        ----------
        name : str
            The channel variable name.
        value : bool
            Whether or not this channel is connected.

        """
        self._connections[name] = value
        self._send_update(self.connected, self._value)

    def calculate_expression(self):
        """
        Evaluate the expression defined by the rule and emit the `rule_signal`
        with the new value.
        """
        vals = self._values.copy()
        if any([vals.get(n) is None for n in self._names]):
            logger.debug('Skipping execution as not all values are set.')
            return

        env = dict(CalcThread.eval_env)
        env.update(**vals)
        env.update({'prev_res': self._value})

        try:
            ret = eval(self._expression, env)
            self._value = ret
            self._send_update(self.connected, ret)
        except Exception as e:
            logger.exception("Error while evaluating CalcPlugin connection %s",
                             self.objectName())


class Connection(PyDMConnection):
    def __init__(self, channel, address, protocol=None, parent=None):
        super(Connection, self).__init__(channel, address, protocol, parent)
        self._calc_thread = None
        self.value = None
        self._configuration = {}
        self._waiting_config = True

        self.add_listener(channel)
        self._init_connection()

    def _init_connection(self):
        self.write_access_signal.emit(False)

    def add_listener(self, channel):
        self._setup_calc(channel)
        super(Connection, self).add_listener(channel)
        self.broadcast_value()

    def broadcast_value(self):
        self.connection_state_signal.emit(self.connected)
        if self.value is not None:
            self.new_value_signal[type(self.value)].emit(self.value)

    def _setup_calc(self, channel):
        if not self._waiting_config:
            logger.debug('CalcPlugin connection already configured.')
            return

        try:
            url_data = UrlToPython(channel)
        except ValueError("Not enough information"):
            logger.debug('Invalid configuration for Calc Plugin connection', exc_info=True)
            return

        self._configuration['name'] = url_data.name
        self._configuration.update(url_data.config)
        self._waiting_config = False

        self._calc_thread = CalcThread(self._configuration)
        self._calc_thread.setObjectName("calc_{}".format(url_data.name))
        self._calc_thread.new_data_signal.connect(self.receive_new_data,
                                                  Qt.QueuedConnection)
        self._calc_thread.start()
        return True

    @Slot(dict)
    def receive_new_data(self, data):
        if not data:
            return
        try:
            conn = data.get('connection')
            self.connected = conn
            self.connection_state_signal.emit(conn)
        except KeyError:
            logger.debug('Connection was not available yet for calc.')
        try:
            val = data.get('value')
            self.value = val
            if val is not None:
                self.new_value_signal[type(val)].emit(val)
        except KeyError:
            logger.debug('Value was not available yet for calc.')

    def close(self):
        self._calc_thread.requestInterruption()


class CalculationPlugin(PyDMPlugin):
    protocol = "calc"
    connection_class = Connection

    @staticmethod
    def get_connection_id(channel):
        obj = UrlToPython(channel)
        return obj.name


class UrlToPython:
    def __init__(self, channel):
        self.channel = channel
        self.address = "calc://" + PyDMPlugin.get_address(self.channel)
        self.name = None
        self.config = None
        self.get_info()

    def get_info(self):
        """
        Parses a given url into a list and a string.

        Returns
        -------
        A tuple: (<list>, <str>)
        """

        try:
            self.config = parse.parse_qs(parse.urlsplit(self.address).query.replace("+", "%2B"))
            self.name = parse.urlsplit(self.address).netloc

            if not self.name or not self.config:
                raise
        except Exception:
            try:
                if not self.name:
                    raise
                logger.debug('Calc Plugin  connection %s got new listener.', self.address)
                return None, self.name, self.address
            except Exeption:
                msg = "Invalid configuration for Calc Plugin  connection. %s"
                logger.exception(msg, self.address, exc_info=True)
                raise ValueError("error in Calc Plugin plugin input")

        return True



import os
import json

from qtpy import QtWidgets, QtCore, QtDesigner
from qtpy.QtWidgets import QApplication, QWidget, QStyle, QStyleOption
from qtpy.QtGui import QPainter, QPixmap
from qtpy.QtCore import Property, Qt, QSize, QSizeF, QRectF, qInstallMessageHandler
from qtpy.QtSvg import QSvgRenderer
from ..utilities import is_pydm_app, is_qt_designer, find_file
from .base import PyDMWidget


class SymbolEditor(QtWidgets.QDialog):
    """
    QDialog for user-friendly editing of the symbols in a widget inside the Qt
    Designer.

    Parameters
    ----------
    widget : PyDMWidget
        The widget which we want to edit the 'imageFiles' property.
    """

    def __init__(self, widget, parent=None):
        super(SymbolEditor, self).__init__(parent)
        
        self.widget = widget
        self.lst_file_item = None
        self.lst_state_item = None
        self.preview = False
        self.preview_file = None
        self.setup_ui()

        try:
            self.symbols = json.loads(widget.imageFiles)
        except:
            self.symbols = {}

        for state, filename in self.symbols.items():
            row = self.tbl_symbols.rowCount()
            self.tbl_symbols.insertRow(row)
            self.tbl_symbols.setItem(row, 0, QtWidgets.QTableWidgetItem(state))
            self.tbl_symbols.setItem(row, 1, QtWidgets.QTableWidgetItem(filename))

    def setup_ui(self):
        """
        Create the required UI elements for the form.

        Returns
        -------
        None
        """

        self.setWindowTitle("PyDM Symbol Widget Editor")
        vlayout = QtWidgets.QVBoxLayout()
        vlayout.setContentsMargins(5, 5, 5, 5)
        vlayout.setSpacing(5)
        self.setLayout(vlayout)

        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setSpacing(5)
        vlayout.addLayout(hlayout)

        # Creating the widgets for the buttons to add and
        # remove symbols
        list_frame = QtWidgets.QFrame(parent=self)
        list_frame.setMinimumHeight(300)
        list_frame.setMinimumWidth(300)
        list_frame.setLineWidth(1)
        list_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        list_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        lf_layout = QtWidgets.QVBoxLayout()
        list_frame.setLayout(lf_layout)

        lf_btn_layout = QtWidgets.QHBoxLayout()
        lf_btn_layout.setContentsMargins(0, 0, 0, 0)
        lf_btn_layout.setSpacing(5)

        self.btn_add_symbol = QtWidgets.QPushButton(parent=self)
        self.btn_add_symbol.setAutoDefault(False)
        self.btn_add_symbol.setDefault(False)
        self.btn_add_symbol.setText("Add Symbol")
        self.btn_add_symbol.clicked.connect(self.add_symbol)

        self.btn_del_symbol = QtWidgets.QPushButton(parent=self)
        self.btn_del_symbol.setAutoDefault(False)
        self.btn_del_symbol.setDefault(False)
        self.btn_del_symbol.setText("Remove Symbol")
        self.btn_del_symbol.clicked.connect(self.del_symbol)

        lf_btn_layout.addWidget(self.btn_add_symbol)
        lf_btn_layout.addWidget(self.btn_del_symbol)

        lf_layout.addLayout(lf_btn_layout)

        # Table containing the state/filename pairs which
        # will display the different symbols
        self.tbl_symbols = QtWidgets.QTableWidget()
        self.tbl_symbols.setShowGrid(True)
        self.tbl_symbols.setCornerButtonEnabled(False)
        headers = ["State", "File"]
        self.tbl_symbols.setColumnCount(len(headers))
        self.tbl_symbols.setHorizontalHeaderLabels(headers)
        header = self.tbl_symbols.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        self.tbl_symbols.itemSelectionChanged.connect(self.load_from_list)
        self.tbl_symbols.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tbl_symbols.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tbl_symbols.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tbl_symbols.verticalHeader().setVisible(False)
        lf_layout.addWidget(self.tbl_symbols)

        hlayout.addWidget(list_frame)

        # Buttons to save or cancel changes made
        buttons_layout = QtWidgets.QHBoxLayout()
        save_btn = QtWidgets.QPushButton("Save", parent=self)
        save_btn.setAutoDefault(False)
        save_btn.setDefault(False)
        save_btn.clicked.connect(self.saveChanges)
        cancel_btn = QtWidgets.QPushButton("Cancel", parent=self)
        cancel_btn.setAutoDefault(False)
        cancel_btn.setDefault(False)
        cancel_btn.clicked.connect(self.cancelChanges)
        buttons_layout.addStretch()
        buttons_layout.addWidget(cancel_btn)
        buttons_layout.addWidget(save_btn)

        vlayout.addLayout(buttons_layout)

        # Creating the widgets that we will use to compose the
        # symbol parameters
        self.frm_edit = QtWidgets.QFrame()
        self.frm_edit.setEnabled(False)
        self.frm_edit.setLineWidth(1)
        self.frm_edit.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_edit.setFrameShape(QtWidgets.QFrame.StyledPanel)

        frm_edit_layout = QtWidgets.QVBoxLayout()
        self.frm_edit.setLayout(frm_edit_layout)

        hlayout.addWidget(self.frm_edit)

        edit_name_layout = QtWidgets.QFormLayout()
        edit_name_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        lbl_state = QtWidgets.QLabel("State:")
        self.txt_state = QtWidgets.QLineEdit()
        self.txt_state.editingFinished.connect(self.state_changed)
        edit_name_layout.addRow(lbl_state, self.txt_state)
        lbl_file = QtWidgets.QLabel("File:")
        self.txt_file = QtWidgets.QLineEdit()
        self.txt_file.textEdited.connect(self.file_changed)
        self.txt_file.returnPressed.connect(self.file_changed)
        edit_name_layout.addRow(lbl_file, self.txt_file)

        self.lbl_image = QtWidgets.QLabel()
        self.lbl_image.setWordWrap(True)
        self.lbl_image.setAlignment(Qt.AlignCenter)
        edit_name_layout.addRow(self.lbl_image)

        frm_edit_layout.addLayout(edit_name_layout)

    def clear_form(self):
        """Clear the form and reset the fields."""
        self.lst_state_item = None
        self.lst_file_item = None
        self.txt_state.setText("")
        self.txt_file.setText("")
        self.lbl_image.setText("")
        self.frm_edit.setEnabled(False)
        self.tbl_symbols.clearSelection()
        self.preview = False

    def load_from_list(self):
        """
        Load an entry from the list into the editing form.

        Returns
        -------
        None
        """
        if not self.tbl_symbols.selectedRanges():
            return

        row = self.tbl_symbols.currentRow()
        self.lst_state_item = self.tbl_symbols.item(row, 0)
        self.lst_file_item = self.tbl_symbols.item(row, 1)
        self.txt_state.setText(self.lst_state_item.text())
        self.txt_file.setText(self.lst_file_item.text())
        self.frm_edit.setEnabled(True)

        filename = self.lst_file_item.text()
        error, self.preview_file = self.check_image(filename)
        if not error:
            self.lbl_image.setText("")
            self.preview = True
        else:
            self.lbl_image.setText(error)
        self.update()

    def add_symbol(self):
        """Add a new rule to the list of rules."""
        default_state = "New State"
        default_file = "New File"

        row = self.tbl_symbols.rowCount()
        self.tbl_symbols.insertRow(row)
        self.lst_state_item = QtWidgets.QTableWidgetItem(default_state)
        self.tbl_symbols.setItem(row, 0, self.lst_state_item)
        self.lst_file_item = QtWidgets.QTableWidgetItem(default_file)
        self.tbl_symbols.setItem(row, 1, self.lst_file_item)

        self.symbols[default_state] = default_file
        self.tbl_symbols.setCurrentItem(self.lst_file_item)
        self.load_from_list()
        self.txt_state.setFocus()

    def del_symbol(self):
        """Delete the selected symbol at the table."""
        if not self.tbl_symbols.selectedRanges():
            return

        confirm_message = "Delete the selected symbol?"
        reply = QtWidgets.QMessageBox().question(self, 'Message',
                                                 confirm_message,
                                                 QtWidgets.QMessageBox.Yes,
                                                 QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            row = self.tbl_symbols.currentRow()
            state_item = self.tbl_symbols.item(row, 0)
            state = state_item.text()
            self.symbols.pop(state, None)
            self.tbl_symbols.removeRow(row)
            self.clear_form()

    def state_changed(self):
        """Callback executed when the state line edit is changed."""
        if not self.lst_state_item:
            return
        prev_state = self.lst_state_item.text()
        self.symbols.pop(prev_state)

        new_state = self.txt_state.text()
        filename = self.txt_file.text()
        self.lst_state_item.setText(new_state)
        self.symbols[new_state] = filename

    def file_changed(self):
        """Callback executed when the file line edit is changed."""
        if not self.lst_file_item:
            return
        state = self.txt_state.text()
        new_filename = self.txt_file.text()
        self.lst_file_item.setText(new_filename)
        self.symbols[state] = new_filename

        error, self.preview_file = self.check_image(new_filename)
        if not error:
            self.lbl_image.setText("")
            self.preview = True
        else:
            self.lbl_image.setText(error)
        self.update()

    def paintEvent(self, event):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        At PyDMSymbolEditor this method handles the image preview.

        Parameters
        ----------
        event : QPaintEvent
        """
        if not self.preview:
            return
        size = QSize(140, 140)
        _painter = QPainter()
        _painter.begin(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, _painter, self)
        image_to_draw = self.preview_file
        if isinstance(image_to_draw, QPixmap):
            w = float(image_to_draw.width())
            h = float(image_to_draw.height())
            sf = min(size.width() / w, size.height() / h)
            scale = (sf, sf)
            _painter.scale(scale[0], scale[1])
            _painter.drawPixmap(335/sf, 120/sf, image_to_draw)
        elif isinstance(image_to_draw, QSvgRenderer):
            draw_size = QSizeF(image_to_draw.defaultSize())
            draw_size.scale(QSizeF(size), Qt.KeepAspectRatio)
            image_to_draw.render(_painter, QRectF(335, 120, draw_size.width(), draw_size.height()))
        _painter.end()
        self.preview = False

    def check_image(self, filename):
        """
        Checks a filename to see if the image can be loaded.
        Parameters
        ----------
        filename : (str)
            Inputted filename by user

        Returns
        -------
        tuple : (str, misc)
            Error message and None if an error is present or None and a
            QSvgRenderer/QPixmap (depending on file type).

        """
        error = None
        file_type = None

        abs_path = os.path.expanduser(os.path.expandvars(filename))

        if not os.path.isabs(abs_path):
            try:
                if is_qt_designer():
                    p = self.get_designer_window()
                    if p is not None:
                        ui_dir = p.absoluteDir().absolutePath()
                        abs_path = os.path.join(ui_dir, abs_path)
                else:
                    parent_display = self.widget.find_parent_display()
                    base_path = None
                    if parent_display:
                        base_path = os.path.dirname(
                            parent_display.loaded_file())
                    abs_path = find_file(abs_path, base_path=base_path)
            except Exception as ex:
                print("Exception: ", ex)
                error = "Unable to find full filepath for {}".format(filename)
                abs_path = filename
        # First, lets try SVG.  We have to try SVG first, otherwise
        # QPixmap will happily load the SVG and turn it into a raster image.
        # Really annoying: We have to try to load the file as SVG,
        # and we expect it will fail often (because many images aren't SVG).
        # Qt prints a warning message to stdout any time SVG loading fails.
        # So we have to temporarily silence Qt warning messages here.
        qInstallMessageHandler(self.qt_message_handler)
        svg = QSvgRenderer()
        if svg.load(abs_path):
            file_type = svg
            qInstallMessageHandler(None)
            return error, file_type
        qInstallMessageHandler(None)
        # SVG didn't work, lets try QPixmap
        image = QPixmap(abs_path)
        if not image.isNull():
            file_type = image
            return error, file_type
        # If we get this far, the file specified could not be loaded at all.
        if error is None:
            error = "Could not load image \n{}".format(filename)
        return error, file_type

    def get_designer_window(self):  # pragma: no cover
        # Internal function to find the designer window that owns this widget.
        p = self.widget.parent()
        while p is not None:
            if isinstance(p, QtDesigner.QDesignerFormWindowInterface):
                return p
            p = p.parent()
        return None

    def is_data_valid(self):
        """
        Sanity check the form data.

        Returns
        -------
        tuple : (bool, str)
            True and "" in case there is no error or False and the error message
            otherwise.
        """
        errors = []
        for state, filename in self.symbols.items():
            if state is None or state == "":
                errors.append("Image {} has no state".format(filename))
            error, file_type = self.check_image(filename)
            if error:
                errors.append(error)

        if len(errors) > 0:
            error_msg = os.linesep.join(errors)
            return False, error_msg

        return True, ""

    def qt_message_handler(self, msg_type, *args):
        # Intentionally suppress all qt messages.  Make sure not to leave this handler installed.
        pass

    @QtCore.Slot()
    def saveChanges(self):
        """Save the new symbols at the widget `symbols` property."""
        # If the form is being edited, we make sure self.symbols has all the
        # latest values from the form before we try to validate.  This fixes
        # a problem where the last form item change wouldn't get saved unless
        # the user knew to hit 'enter' or leave the field to end editing before
        # hitting save.
        if self.frm_edit.isEnabled():
            self.state_changed()
        status, message = self.is_data_valid()
        if status:
            data = json.dumps(self.symbols)
            formWindow = QtDesigner.QDesignerFormWindowInterface.findFormWindow(self.widget)
            if formWindow:
                formWindow.cursor().setProperty("imageFiles", data)
            self.accept()
        else:
            QtWidgets.QMessageBox.critical(self, "Error Saving", message,
                                           QtWidgets.QMessageBox.Ok)

    @QtCore.Slot()
    def cancelChanges(self):
        """Abort the changes and close the dialog."""
        self.close()



import pytest
import logging

from pydm.widgets.timeplot import PyDMTimePlot
from pydm.widgets.waveformplot import WaveformCurveItem
from qtpy.QtGui import QColor
from qtpy.QtCore import QTimer, Qt

from collections import OrderedDict
from ...widgets.baseplot import BasePlotCurveItem, BasePlot

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("color, line_style, line_width, name", [
    (QColor("red"), Qt.SolidLine, 1, "test_name"),
    (None, Qt.DashLine, 10, ""),
    (None, None, None, None)
])
def test_baseplotcurveitem_construct(qtbot, color, line_style, line_width, name):
    base_plotcurve_item = BasePlotCurveItem(color, line_style, line_width, name=name)

    assert base_plotcurve_item._color == color if color else QColor("white")
    assert base_plotcurve_item._pen.color() == color if color else QColor("white")
    assert base_plotcurve_item._pen.style() == line_style if line_style else Qt.SolidLine
    assert base_plotcurve_item._pen.width() == line_width if line_width else 1
    assert base_plotcurve_item.opts["name"] == name


def test_baseplotcurveitem_properties_and_setters(qtbot):
    base_plotcurve_item = BasePlotCurveItem()

    assert base_plotcurve_item.color_string == "white"

    base_plotcurve_item.color_string = "blue"
    assert base_plotcurve_item.color_string == "blue"

    base_plotcurve_item.color = "green"
    assert base_plotcurve_item.color == QColor("green")

    base_plotcurve_item.color = QColor("orange")
    assert base_plotcurve_item.color == QColor("orange")
    assert base_plotcurve_item.lineStyle == Qt.SolidLine

    base_plotcurve_item.lineStyle = Qt.DashLine
    assert base_plotcurve_item.lineStyle == Qt.DashLine
    assert base_plotcurve_item.lineWidth == 1

    base_plotcurve_item.lineWidth = 3
    assert base_plotcurve_item.lineWidth == 3

    base_plotcurve_item.symbol = 'o'
    assert base_plotcurve_item.symbol == 'o'
    assert base_plotcurve_item._pen.color() == base_plotcurve_item._color

    # The invalid symbol 'A' should be ejected
    base_plotcurve_item.symbol = "A"
    assert base_plotcurve_item.symbol == 'o'
    base_plotcurve_item.symbolSize = 100
    assert base_plotcurve_item.symbolSize == 100


def test_baseplotcurveitem_to_dict(qtbot):
    base_plotcurve_item = BasePlotCurveItem()
    dictionary = base_plotcurve_item.to_dict()

    assert isinstance(dictionary, OrderedDict)
    assert dictionary["name"] == base_plotcurve_item.name()
    assert dictionary["color"] == base_plotcurve_item.color_string
    assert dictionary["lineStyle"] == base_plotcurve_item.lineStyle
    assert dictionary["lineWidth"] == base_plotcurve_item.lineWidth
    assert dictionary["symbol"] == base_plotcurve_item.symbol
    assert dictionary["symbolSize"] == base_plotcurve_item.symbolSize


def test_baseplot_construct(qtbot):
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)

    assert base_plot.plotItem == base_plot.getPlotItem()
    assert base_plot.plotItem.buttonsHidden is True
    assert base_plot.getAutoRangeX() is True
    assert base_plot.getAutoRangeY() is True
    assert base_plot.getShowXGrid() is False
    assert base_plot.getShowYGrid() is False
    assert isinstance(base_plot.redraw_timer, QTimer)
    assert base_plot._redraw_rate == 30
    assert base_plot.maxRedrawRate == base_plot._redraw_rate
    assert len(base_plot._curves) == 0
    assert base_plot._title is None
    assert base_plot._show_legend is False
    assert base_plot._legend is not None
    assert base_plot._legend.isVisible() is False


def test_baseplot_add_curve(qtbot):
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)


def test_baseplot_multiple_y_axes(qtbot):
    """ Test that when we add curves while specifying new y-axis names for them, those axes are created and
        added to the plot correctly. Also confirm that adding a curve to an existing axis works properly. """
    base_plot = BasePlot()
    base_plot.clear()
    base_plot.clearAxes()

    # Here we add 4 curves to our plot. Because we use 3 unique axis names, we should see that 3 axes are created. The
    # 4th curve is added with an axis name that already exists, so it should get assigned to that axis rather than
    # have a new one created for it.
    plot_curve_item_1 = WaveformCurveItem()
    plot_curve_item_2 = WaveformCurveItem()
    plot_curve_item_3 = WaveformCurveItem()
    plot_curve_item_4 = WaveformCurveItem()
    base_plot.addCurve(plot_curve_item_1, y_axis_name='Test Axis 1')
    base_plot.addCurve(plot_curve_item_2, y_axis_name='Test Axis 2')
    base_plot.addCurve(plot_curve_item_3, y_axis_name='Test Axis 3')
    base_plot.addCurve(plot_curve_item_4, y_axis_name='Test Axis 1')
    qtbot.addWidget(base_plot)

    # There should be 4 axes (the x-axis, and the 3 new y-axes we have just created)
    assert len(base_plot.plotItem.axes) == 4

    # Verify the 4 axes are indeed the ones we expect
    assert 'bottom' in base_plot.plotItem.axes
    assert 'Test Axis 1' in base_plot.plotItem.axes
    assert 'Test Axis 2' in base_plot.plotItem.axes
    assert 'Test Axis 3' in base_plot.plotItem.axes

    # Verify their orientations were set correctly
    assert base_plot.plotItem.axes['Test Axis 1']['item'].orientation == 'left'
    assert base_plot.plotItem.axes['Test Axis 2']['item'].orientation == 'left'
    assert base_plot.plotItem.axes['Test Axis 3']['item'].orientation == 'left'

    # Verify the curves got assigned to the correct axes
    assert base_plot.plotItem.curvesPerAxis['Test Axis 1'] == 2
    assert base_plot.plotItem.curvesPerAxis['Test Axis 2'] == 1
    assert base_plot.plotItem.curvesPerAxis['Test Axis 3'] == 1


def test_baseplot_no_added_y_axes(qtbot):
    """ Confirm that if the user does not name or create any new y-axes, the plot will still work just fine """
    base_plot = BasePlot()
    base_plot.clear()
    base_plot.clearAxes()

    # Add 3 curves to our plot, but don't bother to use any of the y-axis parameters
    # in addCurve() leaving them to their default of None
    plot_curve_item_1 = WaveformCurveItem()
    plot_curve_item_2 = WaveformCurveItem()
    plot_curve_item_3 = WaveformCurveItem()
    base_plot.addCurve(plot_curve_item_1)
    base_plot.addCurve(plot_curve_item_2)
    base_plot.addCurve(plot_curve_item_3)
    qtbot.addWidget(base_plot)

    # There should be 4 axes, the default ones for PyQtGraph (left, bottom, and the mirrored top and right)
    assert len(base_plot.plotItem.axes) == 4

    # Verify their names were not changed in any way
    assert 'bottom' in base_plot.plotItem.axes
    assert 'left' in base_plot.plotItem.axes
    assert 'top' in base_plot.plotItem.axes
    assert 'right' in base_plot.plotItem.axes

    # Verify their orientations were not changed in any way
    assert base_plot.plotItem.axes['bottom']['item'].orientation == 'bottom'
    assert base_plot.plotItem.axes['left']['item'].orientation == 'left'
    assert base_plot.plotItem.axes['top']['item'].orientation == 'top'
    assert base_plot.plotItem.axes['right']['item'].orientation == 'right'


def test_timeplot_add_multiple_axes(qtbot):
    """ Simliar to the multiple y axes test above, but this one creates the new axes first, and invokes the setters
        for the curves and axes directly, which is exactly what would happen in the flow initiated from designer.
        Since the base plot has no method for setting curves itself, we instead use a time plot. """
    time_plot = PyDMTimePlot()

    # A list of axes represented in json that would be auto-generated by a user creating a new plot
    # We have 4 unique axes, 2 on the left side of the plot and 2 on the right
    json_axes = ['{"name": "Axis 1", "orientation": "left", "minRange": -1.0, "maxRange": 1.0, "autoRange": true}',
                 '{"name": "Axis 2", "orientation": "right", "minRange": -1.0, "maxRange": 1.0, "autoRange": true}',
                 '{"name": "Axis 3", "orientation": "right", "minRange": -0.5, "maxRange": 0.5, "autoRange": false}',
                 '{"name": "Axis 4", "orientation": "left", "minRange": -1.0, "maxRange": 1.0, "autoRange": true}']

    time_plot.setYAxes(json_axes)

    # There should be 5 axes, the x-axis and the 4 we have just created
    assert len(time_plot.plotItem.axes) == 5

    # Verify the 5 axes are indeed the ones we expect
    assert 'bottom' in time_plot.plotItem.axes
    assert 'Axis 1' in time_plot.plotItem.axes
    assert 'Axis 2' in time_plot.plotItem.axes
    assert 'Axis 3' in time_plot.plotItem.axes
    assert 'Axis 4' in time_plot.plotItem.axes

    # Verify their orientations were set correctly
    assert time_plot.plotItem.axes['bottom']['item'].orientation == 'bottom'
    assert time_plot.plotItem.axes['Axis 1']['item'].orientation == 'left'
    assert time_plot.plotItem.axes['Axis 2']['item'].orientation == 'right'
    assert time_plot.plotItem.axes['Axis 3']['item'].orientation == 'right'
    assert time_plot.plotItem.axes['Axis 4']['item'].orientation == 'left'

    # Also confirm that setting a specific range on an axis works
    assert time_plot.plotItem.axes['Axis 3']['item'].min_range == -0.5
    assert time_plot.plotItem.axes['Axis 3']['item'].max_range == 0.5
    assert time_plot.plotItem.axes['Axis 3']['item'].auto_range is False

    # Now let's connect 5 curves with these axes. Sine and cosine will share an axis, the rest will get their own
    json_curves = ['{"channel": "ca://MTEST:SinVal", "name": "Sine", "color": "white", "lineStyle": 1, '
                   '"lineWidth": 1, "symbol": null, "symbolSize": 10, "yAxisName": "Axis 1"}',
                   '{"channel": "ca://MTEST:CosVal", "name": "Cosine", "color": "red", "lineStyle": 1, '
                   '"lineWidth": 1, "symbol": null, "symbolSize": 10, "yAxisName": "Axis 1"}',
                   '{"channel": "ca://MTEST:MeanValue", "name": "Orange Value", "color": "orange", "lineStyle": 1, '
                   '"lineWidth": 1, "symbol": null, "symbolSize": 10, "yAxisName": "Axis 2"}',
                   '{"channel": "ca://MTEST:MaxValue", "name": "Green Value", "color": "forestgreen", "lineStyle": 1, '
                   '"lineWidth": 1, "symbol": null, "symbolSize": 10, "yAxisName": "Axis 3"}',
                   '{"channel": "ca://MTEST:MinValue", "name": "Yellow Value", "color": "yellow", "lineStyle": 1, '
                   '"lineWidth": 1, "symbol": null, "symbolSize": 10, "yAxisName": "Axis 4"}']
    time_plot.setCurves(json_curves)

    # Verify the curves got assigned to the correct axes
    assert len(time_plot.curves) == 5
    assert time_plot.plotItem.curvesPerAxis['Axis 1'] == 2
    assert time_plot.plotItem.curvesPerAxis['Axis 2'] == 1
    assert time_plot.plotItem.curvesPerAxis['Axis 3'] == 1
    assert time_plot.plotItem.curvesPerAxis['Axis 4'] == 1


def test_multiaxis_plot_no_designer_flow(qtbot):
    """ Similar to the tests above except don't use the qt designer flow at all. Verify the correct axes get
        added to the plot in this case as well. """

    # Setup a plot with three test data channels, 2 assigned to the same axis, the third to its own
    plot = PyDMTimePlot()
    plot.addYChannel(y_channel='ca://test_channel', yAxisName='Axis 1')
    plot.addYChannel(y_channel='ca://test_channel_2', yAxisName='Axis 1')
    plot.addYChannel(y_channel='ca://test_channel_3', yAxisName='Axis 2')

    # There should be 5 axes, the 3 we definitely expect ('Axis 1' 'Axis 2', and the default 'bottom' x-axis). But
    # pyqtgraph also adds a mirrored 'right' axis and a mirrored 'top' axis. These do not display by
    # default so we'll keep them.
    assert len(plot.plotItem.axes) == 5
    assert 'bottom' in plot.plotItem.axes
    assert 'Axis 1' in plot.plotItem.axes
    assert 'Axis 2' in plot.plotItem.axes
    assert 'right' in plot.plotItem.axes
    assert 'top' in plot.plotItem.axes
    assert plot.plotItem.curvesPerAxis['Axis 1'] == 2
    assert plot.plotItem.curvesPerAxis['Axis 2'] == 1

    # Now check the case where no new y-axis name is specified for any of the new channels.
    plot = PyDMTimePlot()
    plot.addYChannel(y_channel='ca://test_channel')
    plot.addYChannel(y_channel='ca://test_channel_2')
    plot.addYChannel(y_channel='ca://test_channel_3')

    # Since no new axes were created, we just have the default ones provided by pyqtgraph
    assert len(plot.plotItem.axes) == 4
    assert 'bottom' in plot.plotItem.axes
    assert 'left' in plot.plotItem.axes
    assert 'right' in plot.plotItem.axes
    assert 'top' in plot.plotItem.axes



import functools
import json
import warnings
from typing import Optional
from qtpy.QtGui import QColor, QBrush
from qtpy.QtCore import Signal, Slot, Property, QTimer, Qt, QEvent, QRect
from qtpy.QtWidgets import QToolTip
from .. import utilities
from pyqtgraph import AxisItem, PlotWidget, PlotDataItem, mkPen, ViewBox, InfiniteLine, SignalProxy
from collections import OrderedDict
from typing import Optional
from .base import PyDMPrimitiveWidget, widget_destroyed
from .multi_axis_plot import MultiAxisPlot


class NoDataError(Exception):
    """NoDataError is raised when a curve tries to perform an operation,
    but does not yet have any data."""
    pass


class BasePlotCurveItem(PlotDataItem):
    """
    BasePlotCurveItem represents a single curve in a plot.

    In addition to the parameters listed below, WaveformCurveItem accepts
    keyword arguments for all plot options that pyqtgraph.PlotDataItem accepts.
    Each subclass of ``BasePlotCurveItem`` should have a class attribute
    `_channels` that lets us know the attribute names where we can find
    PyDMChannel objects. This allows us to connect and disconnect these
    connections when appropriate

    Parameters
    ----------
    color : QColor, optional
        The color used to draw the curve line and the symbols.
    lineStyle: int, optional
        Style of the line connecting the data points.
        Must be a value from the Qt::PenStyle enum
        (see http://doc.qt.io/qt-5/qt.html#PenStyle-enum).
    lineWidth: int, optional
        Width of the line connecting the data points.
    yAxisName: str, optional
        The name of the axis to link this curve with. Leaving it None will result in the default
        name of 'Axis 1' which may still be modified later if needed.
    **kargs: optional
        PlotDataItem keyword arguments, such as symbol and symbolSize.
    """

    REDRAW_ON_X, REDRAW_ON_Y, REDRAW_ON_EITHER, REDRAW_ON_BOTH = range(4)
    symbols = OrderedDict([('None', None),
                           ('Circle', 'o'),
                           ('Square', 's'),
                           ('Triangle', 't'),
                           ('Star', 'star'),
                           ('Pentagon', 'p'),
                           ('Hexagon', 'h'),
                           ('X', 'x'),
                           ('Diamond', 'd'),
                           ('Plus', '+')])
    lines = OrderedDict([('NoLine', Qt.NoPen),
                         ('Solid', Qt.SolidLine),
                         ('Dash', Qt.DashLine),
                         ('Dot', Qt.DotLine),
                         ('DashDot', Qt.DashDotLine),
                         ('DashDotDot', Qt.DashDotDotLine)])

    data_changed = Signal()

    def __init__(self, color=None, lineStyle=None, lineWidth=None, yAxisName=None, **kws):
        self._color = QColor('white')
        self._thresholdColor = QColor('white')
        self._pen = mkPen(self._color)
        if lineWidth is not None:
            self._pen.setWidth(lineWidth)
        if lineStyle is not None:
            self._pen.setStyle(lineStyle)
        kws['pen'] = self._pen
        super(BasePlotCurveItem, self).__init__(**kws)
        self.setSymbolBrush(None)
        if color is not None:
            self.color = color

        if yAxisName is None:
            self._y_axis_name = 'Axis 1'
        else:
            self._y_axis_name = yAxisName

        # Bar related items will only be set if the user wants to render the plot as a bar graph
        self.bar_width = None
        # Value above or below these thresholds will be drawn in the threshold color on bar graphs
        self.upper_threshold = None
        self.lower_threshold = None
        self.bar_graph_item = None

        if hasattr(self, "channels"):
            self.destroyed.connect(functools.partial(widget_destroyed,
                                                     self.channels))

    @property
    def color_string(self):
        """
        A string representation of the color used for the curve.  This string
        will be a hex color code, like #FF00FF, or an SVG spec color name, if
        a name exists for the color.

        Returns
        -------
        str
        """
        return str(utilities.colors.svg_color_from_hex(self.color.name(),
                                                       hex_on_fail=True))

    @color_string.setter
    def color_string(self, new_color_string):
        """
        A string representation of the color used for the curve.  This string
        will be a hex color code, like #FF00FF, or an SVG spec color name, if
        a name exists for the color.

        Parameters
        -------
        new_color_string: int
            The new string to use for the curve color.
        """
        self.color = QColor(str(new_color_string))

    @property
    def color(self):
        """
        The color used for the curve.

        Returns
        -------
        QColor
        """
        return self._color

    @color.setter
    def color(self, new_color):
        """
        The color used for the curve.

        Parameters
        -------
        new_color: QColor or str
            The new color to use for the curve.
            Strings are passed to WaveformCurveItem.color_string.
        """
        if isinstance(new_color, str):
            self.color_string = new_color
            return
        self._color = new_color
        self._pen.setColor(self._color)
        self.setPen(self._pen)
        self.setSymbolPen(self._color)

    @property
    def threshold_color_string(self) -> str:
        """
        A string representation of the threshold color used for the bar graph.  This string
        will be a hex color code, like #FF00FF, or an SVG spec color name, if
        a name exists for the color.

        Returns
        -------
        str
        """
        return str(utilities.colors.svg_color_from_hex(self.threshold_color.name(),
                                                       hex_on_fail=True))

    @property
    def threshold_color(self) -> QColor:
        """
        The color used for bars exceeding either the upper or lower thresholds.

        Returns
        -------
        QColor
        """
        return self._thresholdColor

    @threshold_color.setter
    def threshold_color(self, new_color: QColor):
        """
        Set the color used for bars exceeding either the upper or lower thresholds.

        Parameters
        -------
        new_color: QColor
        """
        self._thresholdColor = new_color

    @property
    def y_axis_name(self):
        """
        Return the name of the y-axis that this curve should be associated with. This allows us to have plots that
        contain multiple y-axes, with each curve assigned to either a unique or shared axis as needed.
        Returns
        -------
        str
        """
        return self._y_axis_name

    @y_axis_name.setter
    def y_axis_name(self, axis_name):
        """
        Set the name of the y-axis that should be associated with this curve.
        Parameters
        ----------
        axis_name: str
        """
        self._y_axis_name = axis_name

    @property
    def lineStyle(self):
        """
        Return the style of the line connecting the data points.
        Must be a value from the Qt::PenStyle enum
        (see http://doc.qt.io/qt-5/qt.html#PenStyle-enum).

        Returns
        -------
        int
        """
        return self._pen.style()

    @lineStyle.setter
    def lineStyle(self, new_style):
        """
        Set the style of the line connecting the data points.
        Must be a value from the Qt::PenStyle enum
        (see http://doc.qt.io/qt-5/qt.html#PenStyle-enum).

        Parameters
        -------
        new_style: int
        """
        if new_style in self.lines.values():
            self._pen.setStyle(new_style)
            self.setPen(self._pen)

    @property
    def lineWidth(self):
        """
        Return the width of the line connecting the data points.

        Returns
        -------
        int
        """
        return self._pen.width()

    @lineWidth.setter
    def lineWidth(self, new_width):
        """
        Set the width of the line connecting the data points.

        Parameters
        -------
        new_width: int
        """
        self._pen.setWidth(int(new_width))
        self.setPen(self._pen)

    @property
    def symbol(self):
        """
        The single-character code for the symbol drawn at each datapoint.

        See the documentation for pyqtgraph.PlotDataItem for possible values.

        Returns
        -------
        str or None
        """
        return self.opts['symbol']

    @symbol.setter
    def symbol(self, new_symbol):
        """
        The single-character code for the symbol drawn at each datapoint.

        See the documentation for pyqtgraph.PlotDataItem for possible values.

        Parameters
        -------
        new_symbol: str or None
        """
        if new_symbol in self.symbols.values():
            self.setSymbol(new_symbol)
            self.setSymbolPen(self._color)

    @property
    def symbolSize(self):
        """
        Return the size of the symbol to represent the data.

        Returns
        -------
        int
        """
        return self.opts['symbolSize']

    @symbolSize.setter
    def symbolSize(self, new_size):
        """
        Set the size of the symbol to represent the data.

        Parameters
        -------
        new_size: int
        """
        self.setSymbolSize(int(new_size))

    def setBarGraphInfo(self, bar_width: Optional[float] = 1.0, upper_threshold: Optional[float] = None,
                        lower_threshold: Optional[float] = None, color: Optional[QColor] = None) -> None:
        """
        Set the attributes associated with displaying a plot as a bar graph. These will only be set
        if the plot is to be rendered as a bar graph. And any or all of them may be omitted even if it
        is a bar graph. Omitting color parameters will result in the plot displaying a uniform color.

        Parameters
        ----------
        bar_width: float, optional
            The width of all bars displayed on the plot
        upper_threshold: float, optional
            Any bar above this value will be drawn in the threshold color
        lower_threshold: float, optional
            Any bar below this value will be drawn in the threshold color.
        color: QColor, optional
            The color to draw bars exceeding either threshold in.
        """
        self.bar_width = bar_width
        self.upper_threshold = upper_threshold
        self.lower_threshold = lower_threshold
        self.threshold_color = color

    def to_dict(self):
        """
        Returns an OrderedDict representation with values for all properties
        needed to recreate this curve.

        Returns
        -------
        OrderedDict
        """
        return OrderedDict([("name", self.name()),
                            ("color", self.color_string),
                            ("lineStyle", self.lineStyle),
                            ("lineWidth", self.lineWidth),
                            ("symbol", self.symbol),
                            ("symbolSize", self.symbolSize),
                            ("yAxisName", self.y_axis_name),
                            ("barWidth", self.bar_width),
                            ("upperThreshold", self.upper_threshold),
                            ("lowerThreshold", self.lower_threshold),
                            ("thresholdColor", self.threshold_color_string)])

    def close(self):
        pass


class BasePlotAxisItem(AxisItem):
    """
    BasePlotAxisItem represents a single axis in a plot.

    Parameters
    ----------
    axisName: str
        The name of the axis
    axisOrientation: str, optional
        The orientation of this axis. The default for this value is 'left'. Must be set to either 'right', 'top',
        'bottom', or 'left'. See: https://pyqtgraph.readthedocs.io/en/latest/graphicsItems/axisitem.html
    label: str, optional
        The label to be displayed along the axis
    axisMinRange: float, optional
        The minimum value to be displayed on this axis
    axisMaxRange: float, optional
        The maximum value to be displayed on this axis
    axisAutoRange: bool, optional
        Whether or not this axis should automatically update its range as it receives new data
    **kws: optional
        Extra arguments for CSS style options for this axis
    """

    axis_orientations = OrderedDict([('Left', 'left'),
                                     ('Right', 'right')])

    def __init__(self, name, orientation='left', label=None, minRange=-1.0,
                 maxRange=1.0, autoRange=True, **kws):
        super(BasePlotAxisItem, self).__init__(orientation, **kws)

        self._name = name
        self._orientation = orientation
        self._label = label
        self._min_range = minRange
        self._max_range = maxRange
        self._auto_range = autoRange

    @property
    def name(self):
        """
        Return the name of the axis

        Returns
        -------
        str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Set the name of the axis

        Parameters
        ----------
        name: str
        """
        self._name = name

    @property
    def orientation(self):
        """
        Return the orientation of the y-axis this curve is associated with. Will be 'left', 'right', 'bottom', or 'top'
        See: https://pyqtgraph.readthedocs.io/en/latest/graphicsItems/axisitem.html

        Returns
        -------
        str
        """
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        """
        Set the orientation of the y-axis this curve is associated with. Must be 'left', 'right', 'bottom', or 'top'

        Parameters
        ----------
        orientation: str
        """
        self._orientation = orientation

    @property
    def label_text(self) -> str:
        """ Return the label to be displayed along this axis. """
        return self._label

    @label_text.setter
    def label_text(self, label: str):
        """ Set the label to be displayed along this axis """
        self._label = label

    @property
    def min_range(self):
        """
        Return the minimum range displayed on this axis

        Returns
        -------
        float
        """
        return self._min_range

    @min_range.setter
    def min_range(self, min_range):
        """
        Set the minimum range for this axis

        Parameters
        ----------
        min_range: float
        """
        self._min_range = min_range

    @property
    def max_range(self):
        """
        Return the maximum range displayed on this axis

        Returns
        -------
        float
        """
        return self._max_range

    @max_range.setter
    def max_range(self, max_range):
        """
        Set the maximum range for this axis

        Parameters
        ----------
        max_range: float
        """
        self._max_range = max_range

    @property
    def auto_range(self):
        """
        Return whether or not this axis should automatically update its range when receiving new data

        Returns
        -------
        bool
        """
        return self._auto_range

    @auto_range.setter
    def auto_range(self, auto_range):
        """
        Set whether or not this axis should automatically update its range when receiving new data

        Parameters
        ----------
        auto_range: bool
        """
        self._auto_range = auto_range

    def to_dict(self):
        """
        Returns an OrderedDict representation with values for all properties
        needed to recreate this axis.

        Returns
        -------
        OrderedDict
        """
        return OrderedDict([("name", self._name),
                            ("orientation", self._orientation),
                            ("label", self._label),
                            ("minRange", self._min_range),
                            ("maxRange", self._max_range),
                            ("autoRange", self._auto_range)])


class BasePlot(PlotWidget, PyDMPrimitiveWidget):
    crosshair_position_updated = Signal(float, float)

    def __init__(self, parent=None, background='default', axisItems=None):
        # First create a custom MultiAxisPlot to pass to the base PlotWidget class to support multiple y axes. Note
        # that this plot will still function just fine in the case the user doesn't need additional y axes.
        plotItem = MultiAxisPlot(axisItems=axisItems)
        if axisItems is None or 'left' not in axisItems:
            # The pyqtgraph PlotItem.setAxisItems() will always add an an AxisItem called left whether you asked
            # it to or not. This will clear it if not specifically requested.
            plotItem.removeAxis('left')
        super(BasePlot, self).__init__(parent=parent, background=background, plotItem=plotItem)

        self.plotItem = plotItem
        self.plotItem.hideButtons()
        self._auto_range_x = None
        self.setAutoRangeX(True)
        self._auto_range_y = None
        self.setAutoRangeY(True)
        self._min_x = 0.0
        self._max_x = 1.0
        self._min_y = 0.0
        self._max_y = 1.0
        self._show_x_grid = None
        self.setShowXGrid(False)
        self._show_y_grid = None
        self.setShowYGrid(False)

        self._show_right_axis = False

        self.redraw_timer = QTimer(self)
        self.redraw_timer.timeout.connect(self.redrawPlot)

        self._redraw_rate = 30  # Redraw at 30 Hz by default.
        self.maxRedrawRate = self._redraw_rate
        self._axes = []
        self._curves = []
        self._x_labels = []
        self._y_labels = []
        self._title = None
        self._show_legend = False
        self._legend = self.addLegend()
        self._legend.hide()

        # Drawing crosshair on the ViewBox
        self.vertical_crosshair_line = None
        self.horizontal_crosshair_line = None
        self.crosshair_movement_proxy = None

        # Mouse mode to 1 button (left button draw rectangle for zoom)
        self.plotItem.getViewBox().setMouseMode(ViewBox.RectMode)

        if self.getAxis('bottom') is not None:
            # Disables unexpected axis tick behavior described here:
            # https://pyqtgraph.readthedocs.io/en/latest/graphicsItems/axisitem.html
            self.getAxis('bottom').enableAutoSIPrefix(False)

        if utilities.is_qt_designer():
            self.installEventFilter(self)

    def eventFilter(self, obj, event):
        ret = super(BasePlot, self).eventFilter(obj, event)
        if utilities.is_qt_designer():
            if event.type() == QEvent.Enter:
                QToolTip.showText(
                    self.mapToGlobal(self.rect().center()),
                    'Edit plot curves via Right-Click and select "Edit Curves..."',
                    self,
                    QRect(0, 0, 200, 100),
                    4000)
        else:
            # Somehow super here is not invoking the PyDMPrimitiveWidget
            # eventFilter
            ret = PyDMPrimitiveWidget.eventFilter(self, obj, event)

        return ret

    def addCurve(self, plot_data_item, curve_color=None, y_axis_name=None):
        """
        Adds a curve to this plot. If the y axis parameters are specified, either link this curve to an existing
        axis if that axis is already part of this plot, or create a new one and link the curve to it.
        Parameters
        ----------
        plot_data_item: BasePlotCurveItem
            The curve to add to this plot
        curve_color: QColor, optional
            The color to draw the curve and axis label in
        y_axis_name: str, optional
            The name of the axis to link the curve with. If this is the first time seeing this name,
            then a new axis will be created for it.
        """

        if curve_color is None:
            curve_color = utilities.colors.default_colors[
                    len(self._curves) % len(utilities.colors.default_colors)]
            plot_data_item.color_string = curve_color

        self._curves.append(plot_data_item)

        if y_axis_name is None:
            # If the user did not name the axis, use the default ones. Note: multiple calls to setAxisItems() are ok
            self.plotItem.setAxisItems()
            self.addItem(plot_data_item)
        elif y_axis_name in self.plotItem.axes:
            # If the user has chosen an axis that already exists for this curve, simply link the data to that axis
            self.plotItem.linkDataToAxis(plot_data_item, y_axis_name)
        else:
            # Otherwise we create a brand new axis for this data
            self.addAxis(plot_data_item, y_axis_name, 'left')
        self.redraw_timer.start()
        # Connect channels
        for chan in plot_data_item.channels():
            if chan:
                chan.connect()

    def addAxis(self, plot_data_item: BasePlotCurveItem, name: str, orientation: str,
                label: Optional[str] = None, min_range: Optional[float] = -1.0,
                max_range: Optional[float] = 1.0, enable_auto_range: Optional[bool] = True):
        """
        Create an AxisItem with the input name and orientation, and add it to this plot.
        Parameters
        ----------
        plot_data_item: BasePlotCurveItem
            The curve that will be linked with this new axis
        name: str
            The name that will be assigned to this axis
        orientation: str
            The orientation of this axis, must be in 'left' or 'right'
        label: str, optional
            The label to be displayed along this axis
        min_range: float
            The minimum range to display on the axis
        max_range: float
            The maximum range to display on the axis
        enable_auto_range: bool
            Whether or not to use autorange for this axis. Min and max range will not be respected
            when data goes out of range if this is set to True

        Raises
        ------
        Exception
            Raised by PyQtGraph if the orientation is not in 'left' or 'right'
        """

        if name in self.plotItem.axes:
            return

        axis = BasePlotAxisItem(name=name, orientation=orientation, label=label, minRange=min_range,
                                maxRange=max_range, autoRange=enable_auto_range)
        axis.setLabel(text=label)
        axis.enableAutoSIPrefix(False)
        self._axes.append(axis)
        # If the x axis is just timestamps, we don't want autorange on the x axis
        setXLink = hasattr(self, '_plot_by_timestamps') and self._plot_by_timestamps
        self.plotItem.addAxis(axis, name=name, plotDataItem=plot_data_item, setXLink=setXLink,
                              enableAutoRangeX=self.getAutoRangeX(), enableAutoRangeY=enable_auto_range,
                              minRange=min_range, maxRange=max_range)

    def removeCurve(self, plot_item):
        if plot_item.y_axis_name in self.plotItem.axes:
            self.plotItem.unlinkDataFromAxis(plot_item.y_axis_name)

        self.removeItem(plot_item)
        self._curves.remove(plot_item)
        if len(self._curves) < 1:
            self.redraw_timer.stop()
        # Disconnect channels
        for chan in plot_item.channels():
            if chan:
                chan.disconnect()

    def removeAxisAtIndex(self, axis_index):
        axis = self._axes[axis_index]
        self.plotItem.removeAxis(axis.name)
        self._axes.remove(axis)

    def removeCurveWithName(self, name):
        for curve in self._curves:
            if curve.name() == name:
                self.removeCurve(curve)

    def removeCurveAtIndex(self, index):
        curve_to_remove = self._curves[index]
        self.removeCurve(curve_to_remove)

    def setCurveAtIndex(self, index, new_curve):
        old_curve = self._curves[index]
        self._curves[index] = new_curve
        # self._legend.addItem(new_curve, new_curve.name())
        self.removeCurve(old_curve)

    def curveAtIndex(self, index):
        return self._curves[index]

    def curves(self):
        return self._curves

    def clear(self):
        legend_items = [label.text for (sample, label) in self._legend.items]
        for item in legend_items:
            self._legend.removeItem(item)
        self.plotItem.clear()
        self._curves = []

    def clearAxes(self):
        """ Clear out any added axes on this plot """
        for axis in self._axes:
            axis.deleteLater()
        self.plotItem.clearAxes()
        self._axes = []

    @Slot()
    def redrawPlot(self):
        pass

    def getShowXGrid(self):
        return self._show_x_grid

    def setShowXGrid(self, value, alpha=None):
        self._show_x_grid = value
        self.showGrid(x=self._show_x_grid, alpha=alpha)

    def resetShowXGrid(self):
        self.setShowXGrid(False)

    showXGrid = Property("bool", getShowXGrid, setShowXGrid, resetShowXGrid)

    def getShowYGrid(self):
        return self._show_y_grid

    def setShowYGrid(self, value, alpha=None):
        self._show_y_grid = value
        self.showGrid(y=self._show_y_grid, alpha=alpha)

    def resetShowYGrid(self):
        self.setShowYGrid(False)

    showYGrid = Property("bool", getShowYGrid, setShowYGrid, resetShowYGrid)

    def getBackgroundColor(self):
        return self.backgroundBrush().color()

    def setBackgroundColor(self, color):
        if self.backgroundBrush().color() != color:
            self.setBackgroundBrush(QBrush(color))

    backgroundColor = Property(QColor, getBackgroundColor, setBackgroundColor)

    def getAxisColor(self):
        return self.getAxis('bottom')._pen.color()

    def setAxisColor(self, color):
        if self.getAxis('bottom')._pen.color() != color:
            self.getAxis('bottom').setPen(color)
            self.getAxis('left').setPen(color)
            self.getAxis('top').setPen(color)
            self.getAxis('right').setPen(color)

    axisColor = Property(QColor, getAxisColor, setAxisColor)

    def getYAxes(self):
        """
        Dump the current list of axes and each axis's settings into a list
        of JSON-formatted strings.

        Returns
        -------
        settings : list
            A list of JSON-formatted strings, each containing a curve's
            settings
        """
        return [json.dumps(axis.to_dict()) for axis in self._axes]

    def setYAxes(self, new_list):
        """
        Add a list of axes into the graph.

        Parameters
        ----------
        new_list : list
            A list of JSON-formatted strings, each contains an axis and its
            settings
        """
        try:
            new_list = [json.loads(str(i)) for i in new_list]
        except ValueError as e:
            print("Error parsing curve json data: {}".format(e))
            return
        self.clearAxes()
        for d in new_list:
            self.addAxis(plot_data_item=None, name=d.get('name'), orientation=d.get('orientation'),
                         label=d.get('label'), min_range=d.get('minRange'), max_range=d.get('maxRange'),
                         enable_auto_range=d.get('autoRange'))

    yAxes = Property("QStringList", getYAxes, setYAxes, designable=False)

    def getBottomAxisLabel(self):
        return self.getAxis('bottom').labelText

    def getShowRightAxis(self):
        """
        Provide whether the right y-axis is being shown.

        Returns : bool
        -------
        True if the graph shows the right y-axis. False if not.

        """
        return self._show_right_axis

    def setShowRightAxis(self, show):
        """
        Set whether the graph should show the right y-axis.

        Parameters
        ----------
        show : bool
            True for showing the right axis; False is for not showing.
        """
        if show:
            self.showAxis("right")
        else:
            self.hideAxis("right")
        self._show_right_axis = show

    showRightAxis = Property("bool", getShowRightAxis, setShowRightAxis)

    def getPlotTitle(self):
        if self._title is None:
            return ""
        return str(self._title)

    def setPlotTitle(self, value):
        self._title = str(value)
        if len(self._title) < 1:
            self._title = None
        self.setTitle(self._title)

    def resetPlotTitle(self):
        self._title = None
        self.setTitle(self._title)

    title = Property(str, getPlotTitle, setPlotTitle, resetPlotTitle)

    def getXLabels(self):
        return self._x_labels

    def setXLabels(self, labels):
        if self._x_labels != labels:
            self._x_labels = labels
            label = ""
            if len(self._x_labels) > 0:
                # Hardcoded for now as we only have one axis
                label = self._x_labels[0]
            self.setLabel("bottom", text=label)

    def resetXLabels(self):
        self._x_labels = []
        self.setLabel("bottom", text="")

    xLabels = Property("QStringList", getXLabels, setXLabels, resetXLabels)

    def getYLabels(self):
        warnings.warn("Y Labels should be retrieved from the AxisItem. See: AxisItem.label or AxisItem.labelText"
                      "Example: self.getAxis('Axis Name').labelText",
                      DeprecationWarning,
                      stacklevel=2)
        return self._y_labels

    def setYLabels(self, labels):
        warnings.warn("Y Labels should now be set on the AxisItem itself. See: AxisItem.setLabel() "
                      "Example: self.getAxis('Axis Name').setLabel('Label Name')",
                      DeprecationWarning,
                      stacklevel=2)
        if self._y_labels != labels:
            self._y_labels = labels
            label = ""
            if len(self._y_labels) > 0:
                # Hardcoded for now as we only have one axis
                label = self._y_labels[0]
            if 'left' in self.plotItem.axes:
                self.setLabel("left", text=label)

    def resetYLabels(self):
        warnings.warn("Y Labels should now be set on the AxisItem itself. See: AxisItem.setLabel() "
                      "Example: self.getAxis('Axis Name').setLabel('')",
                      DeprecationWarning,
                      stacklevel=2)
        self._y_labels = []
        self.setLabel("left", text="")

    def getShowLegend(self):
        """
        Check if the legend is being shown.

        Returns : bool
        -------
            True if the legend is displayed on the graph; False if not.
        """
        return self._show_legend

    def setShowLegend(self, value):
        """
        Set to display the legend on the graph.

        Parameters
        ----------
        value : bool
            True to display the legend; False is not.
        """
        self._show_legend = value
        if self._show_legend:
            if self._legend is None:
                self._legend = self.addLegend()
            else:
                self._legend.show()
        else:
            if self._legend is not None:
                self._legend.hide()

    def resetShowLegend(self):
        """
        Reset the legend display status to hidden.
        """
        self.setShowLegend(False)

    showLegend = Property(bool, getShowLegend, setShowLegend, resetShowLegend)

    def getAutoRangeX(self):
        return self._auto_range_x

    def setAutoRangeX(self, value):
        self._auto_range_x = value
        if self._auto_range_x:
            self.plotItem.enableAutoRange(ViewBox.XAxis, enable=self._auto_range_x)

    def resetAutoRangeX(self):
        self.setAutoRangeX(True)

    def getAutoRangeY(self):
        return self._auto_range_y

    def setAutoRangeY(self, value):
        self._auto_range_y = value
        if self._auto_range_y:
            self.plotItem.enableAutoRange(ViewBox.YAxis, enable=self._auto_range_y)

    def resetAutoRangeY(self):
        self.setAutoRangeY(True)

    def getMinXRange(self):
        """
        Minimum X-axis value visible on the plot.

        Returns
        -------
        float
        """
        return self.plotItem.viewRange()[0][0]

    def setMinXRange(self, new_min_x_range):
        """
        Set the minimum X-axis value visible on the plot.

        Parameters
        -------
        new_min_x_range : float
        """
        if self._auto_range_x:
            return
        self._min_x = new_min_x_range
        self.plotItem.setXRange(self._min_x, self._max_x, padding=0)

    def getMaxXRange(self):
        """
        Maximum X-axis value visible on the plot.

        Returns
        -------
        float
        """
        return self.plotItem.viewRange()[0][1]

    def setMaxXRange(self, new_max_x_range):
        """
        Set the Maximum X-axis value visible on the plot.

        Parameters
        -------
        new_max_x_range : float
        """
        if self._auto_range_x:
            return

        self._max_x = new_max_x_range
        self.plotItem.setXRange(self._min_x, self._max_x, padding=0)

    def getMinYRange(self):
        """
        Minimum Y-axis value visible on the plot.

        Returns
        -------
        float
        """
        return self.plotItem.viewRange()[1][0]

    def setMinYRange(self, new_min_y_range):
        """
        Set the minimum Y-axis value visible on the plot.

        Parameters
        -------
        new_min_y_range : float
        """
        if self._auto_range_y:
            return

        self._min_y = new_min_y_range
        self.plotItem.setYRange(self._min_y, self._max_y, padding=0)

    def getMaxYRange(self):
        """
        Maximum Y-axis value visible on the plot.

        Returns
        -------
        float
        """
        return self.plotItem.viewRange()[1][1]

    def setMaxYRange(self, new_max_y_range):
        """
        Set the maximum Y-axis value visible on the plot.

        Parameters
        -------
        new_max_y_range : float
        """
        if self._auto_range_y:
            return

        self._max_y = new_max_y_range
        self.plotItem.setYRange(self._min_y, self._max_y, padding=0)

    @Property(bool)
    def mouseEnabledX(self):
        """
        Whether or not mouse interactions are enabled for the X-axis.

        Returns
        -------
        bool
        """
        return self.plotItem.getViewBox().state['mouseEnabled'][0]

    @mouseEnabledX.setter
    def mouseEnabledX(self, x_enabled):
        """
        Whether or not mouse interactions are enabled for the X-axis.

        Parameters
        -------
        x_enabled : bool
        """
        self.plotItem.setMouseEnabled(x=x_enabled)

    @Property(bool)
    def mouseEnabledY(self):
        """
        Whether or not mouse interactions are enabled for the Y-axis.

        Returns
        -------
        bool
        """
        return self.plotItem.getViewBox().state['mouseEnabled'][1]

    @mouseEnabledY.setter
    def mouseEnabledY(self, y_enabled):
        """
        Whether or not mouse interactions are enabled for the Y-axis.

        Parameters
        -------
        y_enabled : bool
        """
        self.plotItem.setMouseEnabled(y=y_enabled)

    @Property(int)
    def maxRedrawRate(self):
        """
        The maximum rate (in Hz) at which the plot will be redrawn.
        The plot will not be redrawn if there is not new data to draw.

        Returns
        -------
        int
        """
        return self._redraw_rate

    @maxRedrawRate.setter
    def maxRedrawRate(self, redraw_rate):
        """
        The maximum rate (in Hz) at which the plot will be redrawn.
        The plot will not be redrawn if there is not new data to draw.

        Parameters
        -------
        redraw_rate : int
        """
        self._redraw_rate = redraw_rate
        self.redraw_timer.setInterval(int((1.0/self._redraw_rate)*1000))

    def pausePlotting(self):
        self.redraw_timer.stop() if self.redraw_timer.isActive() else self.redraw_timer.start()
        return self.redraw_timer.isActive()

    def mouseMoved(self, evt):
        """
        A handler for the crosshair feature. Every time the mouse move, the mouse coordinates are updated, and the
        horizontal and vertical hairlines will be redrawn at the new coordinate. If a PyDMDisplay object is available,
        that display will also have the x- and y- values to update on the UI.

        Parameters
        -------
        evt: MouseEvent
            The mouse event type, from which the mouse coordinates are obtained.
        """
        pos = evt[0]
        if self.sceneBoundingRect().contains(pos):
            mouse_point = self.getViewBox().mapSceneToView(pos)
            self.vertical_crosshair_line.setPos(mouse_point.x())
            self.horizontal_crosshair_line.setPos(mouse_point.y())

            self.crosshair_position_updated.emit(mouse_point.x(), mouse_point.y())

    def enableCrosshair(self, is_enabled, starting_x_pos, starting_y_pos,  vertical_angle=90, horizontal_angle=0,
                        vertical_movable=False, horizontal_movable=False):
        """
        Enable the crosshair to be drawn on the ViewBox.

        Parameters
        ----------
        is_enabled : bool
            True is to draw the crosshair, False is to not draw.
        starting_x_pos : float
            The x coordinate where to start the vertical crosshair line.
        starting_y_pos : float
            The y coordinate where to start the horizontal crosshair line.
        vertical_angle : float
            The angle to tilt the vertical crosshair line. Default at 90 degrees.
        horizontal_angle
            The angle to tilt the horizontal crosshair line. Default at 0 degrees.
        vertical_movable : bool
            True if the vertical line can be moved by the user; False is not.
        horizontal_movable
            False if the horizontal line can be moved by the user; False is not.
        """
        if is_enabled:
            self.vertical_crosshair_line = InfiniteLine(pos=starting_x_pos, angle=vertical_angle,
                                                        movable=vertical_movable)
            self.horizontal_crosshair_line = InfiniteLine(pos=starting_y_pos, angle=horizontal_angle,
                                                          movable=horizontal_movable)

            self.plotItem.addItem(self.vertical_crosshair_line)
            self.plotItem.addItem(self.horizontal_crosshair_line)
            self.crosshair_movement_proxy = SignalProxy(self.plotItem.scene().sigMouseMoved, rateLimit=60,
                                                        slot=self.mouseMoved)
        else:
            if self.vertical_crosshair_line:
                self.plotItem.removeItem(self.vertical_crosshair_line)
            if self.horizontal_crosshair_line:
                self.plotItem.removeItem(self.horizontal_crosshair_line)
            if self.crosshair_movement_proxy:
                # self.crosshair_movement_proxy.disconnect()
                proxy = self.crosshair_movement_proxy
                proxy.block = True
                try:
                    proxy.signal.disconnect(proxy.signalReceived)
                    proxy.sigDelayed.disconnect(proxy.slot)
                except:
                    pass



import pytest

from ...utilities import colors


def test_read_file():
    assert(colors.svg_color_to_hex_map is not None)
    assert (colors.hex_to_svg_color_map is not None)


def test_svg_color_from_hex():
    svg = colors.svg_color_from_hex('#000000', hex_on_fail=False)
    assert (svg == 'black')
    svg = colors.svg_color_from_hex('#000000', hex_on_fail=True)
    assert (svg == 'black')

    with pytest.raises(KeyError):
        colors.svg_color_from_hex('#XXXXXXX', hex_on_fail=False)
    svg = colors.svg_color_from_hex('#XXXXXXX', hex_on_fail=True)
    assert(svg == '#XXXXXXX')


def test_hex_from_svg_color():
    hex = colors.hex_from_svg_color('black')
    assert (hex == '#000000')

    with pytest.raises(KeyError):
        colors.hex_from_svg_color('invalid_color')



import numpy as np
import pytest
from qtpy.QtCore import Slot

from ..conftest import ConnectionSignals
from ...widgets.archiver_time_plot import ArchivePlotCurveItem, PyDMArchiverTimePlot


@pytest.mark.parametrize('address', ['ca://LINAC:PV1', 'pva://LINAC:PV1', 'LINAC:PV1'])
def test_set_archive_channel(address):
    """ Verify the address for the archiver data plugin is set correctly for all possible EPICS address prefixes """
    curve_item = ArchivePlotCurveItem(channel_address=address)
    assert curve_item.archive_channel.address == 'archiver://pv=LINAC:PV1'


def test_receive_archive_data(signals: ConnectionSignals):
    """ Ensure data from archiver appliance is inserted into the archive buffer correctly """
    curve_item = ArchivePlotCurveItem()
    curve_item.setBufferSize(20)
    curve_item.setArchiveBufferSize(20)

    # We start with a data buffer with some sample values like those generated by a running time plot
    # The x-values are timestamps and the y-values are samples from a PV
    example_live_data = np.array([[100, 101, 102, 103, 104, 105],
                                 [2.15, 2.20, 2.25, 2.22, 2.20, 2.18]])
    starting_data = np.concatenate((np.zeros((2, curve_item.getBufferSize()-6)), example_live_data), axis=1)
    curve_item.points_accumulated += 6
    curve_item.data_buffer = starting_data

    # A quick check to make sure we're starting off correctly
    assert (2, 20) == curve_item.data_buffer.shape

    signals.new_value_signal[np.ndarray].connect(curve_item.receiveArchiveData)

    # First test the most basic case, where we've requested a bit of archive data right before the live data
    mock_archive_data = np.array([[70, 75, 80, 85, 90, 95],
                                  [2.05, 2.08, 2.07, 2.08, 2.12, 2.14]])
    signals.new_value_signal[np.ndarray].emit(mock_archive_data)

    expected_data = np.zeros((2, 14))
    expected_data = np.concatenate((expected_data, mock_archive_data), axis=1)

    # Confirm the archive data was inserted as expected
    assert np.array_equal(curve_item.archive_data_buffer, expected_data)


def test_insert_archive_data():
    """ When first receiving large amounts of data from the archiver appliance, it will be of the 'optimized' form
        in which it is sampled across a fixed number of bins. Drawing a zoom box in this data will get more detailed
        data which must be inserted into the archive data buffer. This tests that insertion is successful. """
    curve_item = ArchivePlotCurveItem()
    curve_item.setBufferSize(10)
    curve_item.setArchiveBufferSize(10)

    # Need some initial data in the live data buffer so that the archive data gets inserted into the proper place
    curve_item.data_buffer = np.array([[130, 140], [8, 9]])
    curve_item.points_accumulated = 2

    # Set up a sample archive buffer
    curve_item.archive_data_buffer = np.array([[0, 0, 0, 0, 100, 105, 110, 115, 120, 125],
                                              [0, 0, 0, 0, 2, 3, 4, 5, 6, 7]],
                                              dtype=float)

    curve_item.archive_points_accumulated = 6
    curve_item.zoomed = True

    # Receive raw data that is more detailed than the two values it will be replacing
    mock_archive_data = np.array([[104, 106, 108, 111],
                                  [2.8, 3.1, 3.7, 3.95]])

    curve_item.insert_archive_data(mock_archive_data)

    # The original average values for timestamps 105 and 106 should now be replace with the actual PV data
    expected_data = np.array([[0, 0, 100, 104, 106, 108, 111, 115, 120, 125],
                             [0, 0, 2, 2.8, 3.1, 3.7, 3.95, 5, 6, 7]])

    assert np.array_equal(curve_item.archive_data_buffer, expected_data)


def test_archive_buffer_full():
    """ If we insert more data points than the archive buffer can hold, then the oldest points are
        removed in favor of the new ones until the user requests further backfill data again """
    curve_item = ArchivePlotCurveItem()
    curve_item.setBufferSize(6)
    curve_item.setArchiveBufferSize(6)
    curve_item.data_buffer = np.array([[130, 140], [8, 9]])
    curve_item.points_accumulated = 2

    # Set up a sample archive buffer that is already full
    curve_item.archive_data_buffer = np.array([[100, 105, 110, 115, 120, 125],
                                              [2, 3, 4, 5, 6, 7]],
                                              dtype=float)
    curve_item.archive_points_accumulated = 6
    curve_item.zoomed = True

    # Receive data that will cause that will not fit in the buffer without deleting other data points
    mock_archive_data = np.array([[104, 106, 108],
                                  [2.8, 3.1, 3.7]])

    curve_item.insert_archive_data(mock_archive_data)

    # This is what is left over after the oldest data points have been trimmed
    expected_data = np.array([[104, 106, 108, 115, 120, 125],
                             [2.8, 3.1, 3.7, 5, 6, 7]])

    assert np.array_equal(curve_item.archive_data_buffer, expected_data)


@Slot(float, float, str)
def inspect_data_request(min_x: float, max_x: float, processing_command: str):
    """ Simple helper function to store the signal parameters it was invoked with """
    inspect_data_request.min_x = min_x
    inspect_data_request.max_x = max_x
    inspect_data_request.processing_command = processing_command


def test_request_data_from_archiver(qtbot):
    """ Test that the signal requesting data from the archiver appliance is built correctly """

    # Create a plot and its associated curve item
    plot = PyDMArchiverTimePlot(optimized_data_bins=10)
    curve_item = ArchivePlotCurveItem()
    # Connect to the helper function above to allow for inspection of the parameters the signal was invoked with
    curve_item.archive_data_request_signal.connect(inspect_data_request)
    plot._curves.append(curve_item)

    # Request data from a short 100 second period
    plot._archive_request_queued = True
    plot.requestDataFromArchiver(100, 200)

    # Verify that the data is requested for the time period specified, and since it is only 100 seconds, it is raw data
    assert inspect_data_request.min_x == 100
    assert inspect_data_request.max_x == 199
    assert inspect_data_request.processing_command == ''

    # Now request over a day's worth of data at once. This will cause the request to be for optimized data
    # returned in 10 bins as specified by the "optimized_data_bins" param above
    plot._archive_request_queued = True
    plot.requestDataFromArchiver(100, 100000)
    assert inspect_data_request.min_x == 100
    assert inspect_data_request.max_x == 99999
    assert inspect_data_request.processing_command == 'optimized_10'

    # Finally let's do a test without specifying min_x and max_x to test the plot's logic of determining
    # these values itself

    # Create a small data buffer for the plot's curve representing live data visible on the plot
    curve_item.points_accumulated = 3
    curve_item._bufferSize = 5
    # Index 0 represents timestamps, index 1 the associated values. So observations were made at time 300, 301, 302.
    curve_item.data_buffer = np.array([[0, 0, 300, 301, 302], [0, 0, 1.5, 1.6, 1.5]])

    plot._min_x = 50  # This is the minimum timestamp visible on the x-axis, representing what the user panned to
    plot._archive_request_queued = True
    plot.requestDataFromArchiver()
    # The min_x requested should have defaulted to 50 since that is what the user requested as mentioned above
    assert inspect_data_request.min_x == 50
    # Because the oldest live timestamp in the data buffer was 300, the ending timestamp for the request should
    # be one less than that.
    assert inspect_data_request.max_x == 299



from typing import Any, Optional
from qtpy.QtCore import QModelIndex, QObject, QVariant
from qtpy.QtGui import QColor
from .archiver_time_plot import ArchivePlotCurveItem
from .baseplot import BasePlot
from .baseplot_table_model import BasePlotCurvesModel
from .baseplot_curve_editor import BasePlotCurveEditorDialog, PlotStyleColumnDelegate


class PyDMArchiverTimePlotCurvesModel(BasePlotCurvesModel):
    """ Model used in designer for editing archiver time plot curves. """

    def __init__(self, plot: BasePlot, parent: Optional[QObject] = None):
        super().__init__(plot, parent=parent)
        self._column_names = ("Channel", "Archive Data") + self._column_names

    def get_data(self, column_name: str, curve: ArchivePlotCurveItem) -> Any:
        """ Get data for the input column name """
        if column_name == "Channel":
            if curve.address is None:
                return QVariant()
            return str(curve.address)
        elif column_name == "Archive Data":
            return bool(curve.use_archive_data)
        return super().get_data(column_name, curve)

    def set_data(self, column_name: str, curve: ArchivePlotCurveItem, value: Any) -> bool:
        """ Set data on the input curve for the given name and value. Return true if successful. """
        if column_name == "Channel":
            curve.address = str(value)
        elif column_name == "Archive Data":
            curve.use_archive_data = bool(value)
        else:
            return super().set_data(column_name=column_name, curve=curve, value=value)
        return True

    def append(self, address: Optional[str] = None, name: Optional[str] = None, color: Optional[QColor] = None) -> None:
        """ Add a row for a curve with the input address """
        self.beginInsertRows(QModelIndex(), len(self._plot._curves), len(self._plot._curves))
        self._plot.addYChannel(address, name, color)
        self.endInsertRows()

    def removeAtIndex(self, index: QModelIndex):
        """ Remove the row at the input index """
        self.beginRemoveRows(QModelIndex(), index.row(), index.row())
        self._plot.removeYChannelAtIndex(index.row())
        self.endRemoveRows()


class ArchiverTimePlotCurveEditorDialog(BasePlotCurveEditorDialog):
    """ ArchiverTimePlotCurveEditorDialog is a QDialog that is used in Qt Designer
    to edit the properties of the curves in a waveform plot.  This dialog is
    shown when you double-click the plot, or when you right click it and
    choose 'edit curves'.

    This thing is mostly just a wrapper for a table view, with a couple
    buttons to add and remove curves, and a button to save the changes."""
    TABLE_MODEL_CLASS = PyDMArchiverTimePlotCurvesModel

    def __init__(self, plot, parent=None):
        super().__init__(plot, parent)

        plot_style_delegate = PlotStyleColumnDelegate(self, self.table_model, self.table_view)
        plot_style_delegate.hideColumns(hide_line_columns=False, hide_bar_columns=True)



import logging
import os

from .qtplugin_base import qtplugin_factory, WidgetCategory
from .qtplugin_extensions import (RulesExtension, WaveformCurveEditorExtension,
                                  TimeCurveEditorExtension, ArchiveTimeCurveEditorExtension,
                                  ScatterCurveEditorExtension, SymbolExtension)
from .tab_bar_qtplugin import TabWidgetPlugin
from .byte import PyDMByteIndicator

from .checkbox import PyDMCheckbox
from .datetime import (PyDMDateTimeEdit, PyDMDateTimeLabel)
from .drawing import (PyDMDrawingLine, PyDMDrawingRectangle,
                      PyDMDrawingTriangle,
                      PyDMDrawingEllipse, PyDMDrawingCircle, PyDMDrawingArc,
                      PyDMDrawingPie, PyDMDrawingChord, PyDMDrawingImage,
                      PyDMDrawingPolygon, PyDMDrawingPolyline)

from .embedded_display import PyDMEmbeddedDisplay
from .enum_button import PyDMEnumButton
from .enum_combo_box import PyDMEnumComboBox
from .frame import PyDMFrame
from .image import PyDMImageView
from .label import PyDMLabel
from .line_edit import PyDMLineEdit
from .logdisplay import PyDMLogDisplay
from .pushbutton import PyDMPushButton
from .related_display_button import PyDMRelatedDisplayButton
from .shell_command import PyDMShellCommand
from .slider import PyDMSlider
from .spinbox import PyDMSpinbox
from .symbol import PyDMSymbol
from .waveformtable import PyDMWaveformTable
from .scale import PyDMScaleIndicator
from .timeplot import PyDMTimePlot
from .archiver_time_plot import PyDMArchiverTimePlot
from .waveformplot import PyDMWaveformPlot
from .scatterplot import PyDMScatterPlot
from .template_repeater import PyDMTemplateRepeater
from .terminator import PyDMTerminator

from ..utilities.iconfont import IconFont

logger = logging.getLogger(__name__)

ifont = IconFont()

BASE_EXTENSIONS = [RulesExtension]

# Label plugin
PyDMLabelPlugin = qtplugin_factory(PyDMLabel, group=WidgetCategory.DISPLAY,
                                   extensions=BASE_EXTENSIONS,
                                   icon=ifont.icon("tag"))

# Time Plot plugin
PyDMTimePlotPlugin = qtplugin_factory(PyDMTimePlot, group=WidgetCategory.PLOT,
                                      extensions=[TimeCurveEditorExtension,
                                                  RulesExtension],
                                      icon=ifont.icon("chart-line"))

# In order to keep the archiver functionality invisible to users who do not have access to an instance of the
# archiver appliance, only load this if the user has the associated environment variable set
if "PYDM_ARCHIVER_URL" in os.environ:
    # Time Plot with archiver appliance support plugin
    PyDMArchiverTimePlotPlugin = qtplugin_factory(PyDMArchiverTimePlot, group=WidgetCategory.PLOT,
                                                  extensions=[ArchiveTimeCurveEditorExtension,
                                                              RulesExtension],
                                                  icon=ifont.icon("chart-line"))

# Waveform Plot plugin
PyDMWaveformPlotPlugin = qtplugin_factory(PyDMWaveformPlot,
                                          group=WidgetCategory.PLOT,
                                          extensions=[
                                              WaveformCurveEditorExtension,
                                              RulesExtension],
                                          icon=ifont.icon("wave-square"))

# Scatter Plot plugin
PyDMScatterPlotPlugin = qtplugin_factory(PyDMScatterPlot,
                                         group=WidgetCategory.PLOT,
                                         extensions=[
                                             ScatterCurveEditorExtension,
                                             RulesExtension],
                                         icon=ifont.icon("project-diagram"))

# Byte plugin
PyDMByteIndicatorPlugin = qtplugin_factory(PyDMByteIndicator,
                                           group=WidgetCategory.DISPLAY,
                                           extensions=BASE_EXTENSIONS,
                                           icon=ifont.icon("ellipsis-v"))

# Checkbox plugin
PyDMCheckboxPlugin = qtplugin_factory(PyDMCheckbox, group=WidgetCategory.INPUT,
                                      extensions=BASE_EXTENSIONS,
                                      icon=ifont.icon("check-square"))

# Date/Time plugins
PyDMDateTimeEditPlugin = qtplugin_factory(PyDMDateTimeEdit,
                                          group=WidgetCategory.INPUT,
                                          extensions=BASE_EXTENSIONS,
                                          icon=ifont.icon("calendar-minus"))

PyDMDateTimeLabelPlugin = qtplugin_factory(PyDMDateTimeLabel,
                                           group=WidgetCategory.DISPLAY,
                                           extensions=BASE_EXTENSIONS,
                                           icon=ifont.icon("calendar-alt"))
# Drawing plugins
PyDMDrawingArcPlugin = qtplugin_factory(PyDMDrawingArc,
                                        group=WidgetCategory.DRAWING,
                                        extensions=BASE_EXTENSIONS,
                                        icon=ifont.icon("circle-notch"))
PyDMDrawingChordPlugin = qtplugin_factory(PyDMDrawingChord,
                                          group=WidgetCategory.DRAWING,
                                          extensions=BASE_EXTENSIONS,
                                          icon=ifont.icon("moon"))
PyDMDrawingCirclePlugin = qtplugin_factory(PyDMDrawingCircle,
                                           group=WidgetCategory.DRAWING,
                                           extensions=BASE_EXTENSIONS,
                                           icon=ifont.icon("circle"))
PyDMDrawingEllipsePlugin = qtplugin_factory(PyDMDrawingEllipse,
                                            group=WidgetCategory.DRAWING,
                                            extensions=BASE_EXTENSIONS,
                                            icon=ifont.icon("ellipsis-h"))
PyDMDrawingImagePlugin = qtplugin_factory(PyDMDrawingImage,
                                          group=WidgetCategory.DRAWING,
                                          extensions=BASE_EXTENSIONS,
                                          icon=ifont.icon("image"))
PyDMDrawingLinePlugin = qtplugin_factory(PyDMDrawingLine,
                                         group=WidgetCategory.DRAWING,
                                         extensions=BASE_EXTENSIONS,
                                         icon=ifont.icon("minus"))
PyDMDrawingPiePlugin = qtplugin_factory(PyDMDrawingPie,
                                        group=WidgetCategory.DRAWING,
                                        extensions=BASE_EXTENSIONS,
                                        icon=ifont.icon("pizza-slice"))

PyDMDrawingRectanglePlugin = qtplugin_factory(PyDMDrawingRectangle,
                                              group=WidgetCategory.DRAWING,
                                              extensions=BASE_EXTENSIONS,
                                              icon=ifont.icon("border-style"))
PyDMDrawingTrianglePlugin = qtplugin_factory(PyDMDrawingTriangle,
                                             group=WidgetCategory.DRAWING,
                                             extensions=BASE_EXTENSIONS,
                                             icon=ifont.icon("caret-up"))

PyDMDrawingPolygonPlugin = qtplugin_factory(PyDMDrawingPolygon,
                                            group=WidgetCategory.DRAWING,
                                            extensions=BASE_EXTENSIONS,
                                            icon=ifont.icon("draw-polygon"))

PyDMDrawingPolylinePlugin = qtplugin_factory(PyDMDrawingPolyline,
                                            group=WidgetCategory.DRAWING,
                                            extensions=BASE_EXTENSIONS,
                                            icon=ifont.icon("share-alt"))

# Embedded Display plugin
PyDMEmbeddedDisplayPlugin = qtplugin_factory(PyDMEmbeddedDisplay,
                                             group=WidgetCategory.CONTAINER,
                                             extensions=BASE_EXTENSIONS,
                                             icon=ifont.icon("layer-group"))

# Enum Button plugin
PyDMEnumButtonPlugin = qtplugin_factory(PyDMEnumButton,
                                        group=WidgetCategory.INPUT,
                                        extensions=BASE_EXTENSIONS,
                                        icon=ifont.icon("bars"))

# Enum Combobox plugin
PyDMEnumComboBoxPlugin = qtplugin_factory(PyDMEnumComboBox,
                                          group=WidgetCategory.INPUT,
                                          extensions=BASE_EXTENSIONS,
                                          icon=ifont.icon("list-ol"))

# Frame plugin
PyDMFramePlugin = qtplugin_factory(PyDMFrame, group=WidgetCategory.CONTAINER,
                                   is_container=True,
                                   extensions=BASE_EXTENSIONS,
                                   icon=ifont.icon("expand"))

# Image plugin
PyDMImageViewPlugin = qtplugin_factory(PyDMImageView,
                                       group=WidgetCategory.DISPLAY,
                                       extensions=BASE_EXTENSIONS,
                                       icon=ifont.icon("camera"))

# Line Edit plugin
PyDMLineEditPlugin = qtplugin_factory(PyDMLineEdit, group=WidgetCategory.INPUT,
                                      extensions=BASE_EXTENSIONS,
                                      icon=ifont.icon("edit"))

# Log Viewer
PyDMLogDisplayPlugin = qtplugin_factory(PyDMLogDisplay,
                                        group=WidgetCategory.DISPLAY,
                                        extensions=BASE_EXTENSIONS,
                                        icon=ifont.icon("clipboard"))

# Push Button plugin
PyDMPushButtonPlugin = qtplugin_factory(PyDMPushButton,
                                        group=WidgetCategory.INPUT,
                                        extensions=BASE_EXTENSIONS,
                                        icon=ifont.icon("mouse"))

# Related Display Button plugin
PyDMRelatedDisplayButtonPlugin = qtplugin_factory(PyDMRelatedDisplayButton,
                                                  group=WidgetCategory.DISPLAY,
                                                  extensions=BASE_EXTENSIONS,
                                                  icon=ifont.icon(
                                                      "window-maximize"))

# Shell Command plugin
PyDMShellCommandPlugin = qtplugin_factory(PyDMShellCommand,
                                          group=WidgetCategory.INPUT,
                                          extensions=BASE_EXTENSIONS,
                                          icon=ifont.icon("terminal"))

# Slider plugin
PyDMSliderPlugin = qtplugin_factory(PyDMSlider, group=WidgetCategory.INPUT,
                                    extensions=BASE_EXTENSIONS,
                                    icon=ifont.icon("sliders-h"))

# Spinbox plugin
PyDMSpinboxplugin = qtplugin_factory(PyDMSpinbox, group=WidgetCategory.INPUT,
                                     extensions=BASE_EXTENSIONS,
                                     icon=ifont.icon("sort-numeric-up"))

# Scale Indicator plugin
PyDMScaleIndicatorPlugin = qtplugin_factory(PyDMScaleIndicator,
                                            group=WidgetCategory.DISPLAY,
                                            extensions=BASE_EXTENSIONS,
                                            icon=ifont.icon("level-up-alt")
                                            )

# Symbol plugin
PyDMSymbolPlugin = qtplugin_factory(PyDMSymbol, group=WidgetCategory.DISPLAY,
                                    extensions=[SymbolExtension,
                                                RulesExtension],
                                    icon=ifont.icon("icons"))

# Waveform Table plugin
PyDMWaveformTablePlugin = qtplugin_factory(PyDMWaveformTable,
                                           group=WidgetCategory.INPUT,
                                           extensions=BASE_EXTENSIONS,
                                           icon=ifont.icon("table"))

# Tab Widget plugin
PyDMTabWidgetPlugin = TabWidgetPlugin(extensions=BASE_EXTENSIONS)

# Template Repeater plugin
PyDMTemplateRepeaterPlugin = qtplugin_factory(PyDMTemplateRepeater,
                                              group=WidgetCategory.CONTAINER,
                                              extensions=BASE_EXTENSIONS,
                                              icon=ifont.icon("align-justify"))

# Terminator Widget plugin
PyDMTerminatorPlugin = qtplugin_factory(PyDMTerminator,
                                        group=WidgetCategory.MISC,
                                        extensions=BASE_EXTENSIONS)



# Unit Tests for the Channel widget class
import pytest

from ...widgets.label import PyDMLabel
from ...widgets.line_edit import PyDMLineEdit
from ...widgets.channel import PyDMChannel
from pydm.data_plugins import plugin_for_address


class A():
    pass

def test_construct(qtbot):
    """
    Test the construct of the widget.

    Expectations:
    1. The widget is created with all the default values for its properties
    2. When created for a specific widget type, the Channel widget contains the default values, i.e. the appropriate
        slots, for that specific widget type
    3. Channel widgets can be compared for equality and inequality using the '==' and '!=' operators.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_channel = PyDMChannel()

    assert pydm_channel.address is None and \
        pydm_channel.connection_slot is None and \
        pydm_channel.value_slot is None and \
        pydm_channel.severity_slot is None and \
        pydm_channel.enum_strings_slot is None and \
        pydm_channel.unit_slot is None and \
        pydm_channel.prec_slot is None and \
        pydm_channel.upper_ctrl_limit_slot is None and \
        pydm_channel.lower_ctrl_limit_slot is None and \
        pydm_channel.write_access_slot is None and \
        pydm_channel.value_signal is None

    pydm_label = PyDMLabel(init_channel='tst://this')
    qtbot.addWidget(pydm_label)

    pydm_label_channels = pydm_label.channels()[0]
    default_pydm_label_channels = PyDMChannel(address=pydm_label.channel,
                                              connection_slot=pydm_label.connectionStateChanged,
                                              value_slot=pydm_label.channelValueChanged,
                                              severity_slot=pydm_label.alarmSeverityChanged,
                                              enum_strings_slot=pydm_label.enumStringsChanged,
                                              unit_slot=pydm_label.unitChanged,
                                              prec_slot=pydm_label.precisionChanged,
                                              upper_ctrl_limit_slot=pydm_label.upperCtrlLimitChanged,
                                              lower_ctrl_limit_slot=pydm_label.lowerCtrlLimitChanged,
                                              value_signal=None,
                                              write_access_slot=None)
    assert pydm_label_channels == default_pydm_label_channels

    pydm_lineedit = PyDMLineEdit(init_channel='tst://this2')
    qtbot.addWidget(pydm_lineedit)

    # Test equal and not equal comparisons
    pydm_lineedit_channels = pydm_lineedit.channels()[0]
    assert pydm_lineedit_channels != default_pydm_label_channels

    not_same_type = A()
    equal_result = not_same_type == default_pydm_label_channels
    not_equal_result = not_same_type != default_pydm_label_channels
    assert equal_result is False and not_equal_result is True


def test_pydm_connection(test_plugin):
    # Plugin, Channel and Registry
    chan = PyDMChannel('tst://Tst:this3')
    plugin = plugin_for_address(chan.address)
    plugin_no = len(plugin.connections)

    print('Connections Before:')
    for k, v in plugin.connections.items():
        print('\t', k, ' - ', v)

    # Make a connection
    chan.connect()

    print('Connections After:')
    for k, v in plugin.connections.items():
        print('\t', k, ' - ', v)

    assert len(plugin.connections) == plugin_no + 1
    # Remove connections
    chan.disconnect()
    assert len(plugin.connections) == plugin_no



@pytest.mark.parametrize(
    "ch, ch_expected", [
        ("ca://MTEST:Float", "ca://MTEST:Float"),
        (" foo://bar", "foo://bar"),
        (" foo://bar ", "foo://bar"),
        ("foo://bar ", "foo://bar"),
        ("\nfoo://bar", "foo://bar"),
        ("\tfoo://bar", "foo://bar"),
        ("", ""),
        (None, None),
    ])
def test_channel_address(ch, ch_expected):
    channel = PyDMChannel()
    channel.address = ch
    assert channel.address == ch_expected
